#!/usr/bin/env python3
# -*- coding: utf-8 -*-
import datetime

import click
from flask.cli import current_app, with_appcontext
from patientMatcher.resources import path_to_benchmark_patients
from patientMatcher.utils.add import add_node, load_demo_patients
from patientMatcher.utils.delete import drop_all_collections


@click.group()
def add():
    """Add items to database using the CLI"""
    pass


@add.command()
@click.option("-i", "--id", type=click.STRING, nargs=1, required=True, help="Server ID")
@click.option("-l", "--label", type=click.STRING, nargs=1, required=True, help="Server Description")
@click.option(
    "-t", "--token", type=click.STRING, nargs=1, required=True, help="Authorization token"
)
@click.option(
    "-m",
    "--matching_url",
    type=click.STRING,
    nargs=1,
    required=True,
    help="URL to send match requests to",
)
@click.option(
    "-a",
    "--accepted_content",
    type=click.STRING,
    nargs=1,
    required=True,
    help="Accepted Content-Type",
    default="application/vnd.ga4gh.matchmaker.v1.0+json",
)
@click.option(
    "-c", "--contact", type=click.STRING, nargs=1, required=False, help="An email address"
)
@with_appcontext
def node(id, label, token, matching_url, accepted_content, contact=None):
    """Adds a new server to database"""
    click.echo("Adding a new MatchMaker node to database")
    node_obj = {
        "_id": id,
        "label": label,
        "created": datetime.datetime.now(),
        "auth_token": token,
        "matching_url": matching_url,
        "accepted_content": accepted_content,
        "contact": contact,
    }
    inserted_id, collection = add_node(mongo_db=current_app.db, obj=node_obj, is_client=False)
    if inserted_id:
        click.echo(
            'Inserted node with ID "{}" into database collection {}'.format(inserted_id, collection)
        )
    else:
        click.echo("Aborted")


@add.command()
@click.option("-i", "--id", type=click.STRING, nargs=1, required=True, help="Client ID")
@click.option(
    "-t", "--token", type=click.STRING, nargs=1, required=True, help="Authorization token"
)
@click.option("-u", "--url", type=click.STRING, nargs=1, required=True, help="Client URL")
@click.option("-c", "--contact", type=click.STRING, nargs=1, required=False, help="Client email")
@with_appcontext
def client(id, token, url, contact=None):
    """Adds a new client to database"""
    click.echo("Adding a new client to database")
    client_obj = {
        "_id": id,
        "created": datetime.datetime.now(),
        "auth_token": token,
        "base_url": url,
        "contact": contact,
    }
    inserted_id, collection = add_node(mongo_db=current_app.db, obj=client_obj, is_client=True)
    if inserted_id:
        click.echo(
            'Inserted client with ID "{}" into database collection {}'.format(
                inserted_id, collection
            )
        )
    else:
        click.echo("Aborted")


@add.command()
@with_appcontext
def demodata():
    """Adds a set of 50 demo patients to database"""
    mongo_db = current_app.db
    drop_all_collections(mongo_db)
    click.echo("Adding 50 test patients to database..")
    inserted_ids = load_demo_patients(
        path_to_json_data=path_to_benchmark_patients,
        mongo_db=mongo_db,
        convert_to_ensembl=True,  # Save Ensembl IDs for demo genes by default
    )
    click.echo("inserted {} patients into db".format(len(inserted_ids)))

    click.echo("Loading demo client: DExter MOrgan with token:DEMO")
    demo_client = dict(
        _id="DExterMOrgan",
        created=datetime.datetime.now(),
        auth_token="DEMO",
        base_url="Demo client URL",
        contact="demo@patientMatcher.se",
    )
    add_node(mongo_db=mongo_db, obj=demo_client, is_client=True)


add.add_command(node)
add.add_command(client)
add.add_command(demodata)



import pkg_resources

###### Files ######
hpo_filename = "resources/hp.obo.txt"
phenotype_annotation_filename = "resources/phenotype_annotation.tab.txt"
benchmark_patients = "resources/benchmark_patients.json"
json_api = "resources/api.json"

###### Paths ######
path_to_hpo_terms = pkg_resources.resource_filename("patientMatcher", hpo_filename)

path_to_phenotype_annotations = pkg_resources.resource_filename(
    "patientMatcher", phenotype_annotation_filename
)

path_to_benchmark_patients = pkg_resources.resource_filename("patientMatcher", benchmark_patients)

path_to_json_api = pkg_resources.resource_filename("patientMatcher", json_api)



try:
    from werkzeug.contrib.fixers import ProxyFix
except ModuleNotFoundError:
    # Werkzeug >1.0 moved the library to middleware : https://github.com/pallets/werkzeug/issues/1477
    from werkzeug.middleware.proxy_fix import ProxyFix

from patientMatcher.server import create_app

app = create_app()

app.wsgi_app = ProxyFix(app.wsgi_app)



# -*- coding: utf-8 -*-
import json

import mongomock
import pytest
from patientMatcher.resources import path_to_benchmark_patients
from patientMatcher.server import create_app

DATABASE = "testdb"

HGNC_SYMBOLS_2_ENSEMBL_IDS = {
    "AAGAB": "ENSG00000103591",
    "LAMA1": "ENSG00000101680",
    "SNRPB": "ENSG00000101680",
    "NGLY1": "ENSG00000151092",
    "LIMS2": "ENSG00000072163",
    "KARS": "ENSG00000065427",
    "GUCY2C": "ENSG00000070019",
    "GPX4": "ENSG00000167468",
    "STAMBP": "ENSG00000124356",
    "HSD17B4": "ENSG00000133835",
    "GARS": "ENSG00000106105",
    "ALDH6A1": "ENSG00000119711",
    "EFTUD2": "ENSG00000108883",
    "SYNJ1": "ENSG00000159082",
    "IARS2": "ENSG00000067704",
    "GRIN2A": "ENSG00000183454",
    "WDR62": "ENSG00000075702",
    "ATP1A3": "ENSG00000272584",
    "SKI": "ENSG00000157933",
    "PIK3R1": "ENSG00000145675",
    "THOC6": "ENSG00000131652",
    "G6PC3": "ENSG00000141349",
    "TTC7A": "ENSG00000068724",
    "SGOL1": "ENSG00000129810",
}


@pytest.fixture
def mock_symbol_2_ensembl():
    return HGNC_SYMBOLS_2_ENSEMBL_IDS


class MockMail:
    _send_was_called = False
    _message = None

    def send(self, message):
        self._send_was_called = True
        self._message = message


@pytest.fixture
def mock_app(database, pymongo_client):
    app = create_app()
    app.db = database
    app.client = pymongo_client
    return app


@pytest.fixture
def mock_mail():
    return MockMail()


@pytest.fixture
def mock_sender():
    return "mock_sender"


@pytest.fixture(scope="function")
def pymongo_client(request):
    """Get a client to the mongo database"""
    mock_client = mongomock.MongoClient()

    def teardown():
        mock_client.drop_database(DATABASE)

    request.addfinalizer(teardown)
    return mock_client


@pytest.fixture(scope="function")
def database(request, pymongo_client):
    """Get an adapter connected to mongo database"""
    mongo_client = pymongo_client
    database = mongo_client[DATABASE]
    return database


@pytest.fixture(scope="function")
def demo_data_path():
    return path_to_benchmark_patients


@pytest.fixture(scope="function")
def test_client():
    """Returns a test client object"""
    client = {
        "_id": "client_1",
        "auth_token": "superSecretToken",
        "base_url": "https://test_node_institution.com",
        "contact": "test_client@email.com",
    }
    return client


@pytest.fixture(scope="function")
def test_node():
    """Returns a test client object"""
    node = {
        "_id": "node_1",
        "label": "Test node description",
        "auth_token": "superSecretToken",
        "matching_url": "https://test_node/match/",
        "accepted_content": "application/vnd.ga4gh.matchmaker.v1.0+json",
        "contact": "test_node_user@email.com",
    }
    return node


@pytest.fixture(scope="function")
def entrez_gene_patient():
    """Returns a test patient with an entrez gene ID"""
    patient = {
        "contact": {
            "href": "mailto:someuser@mail.com",
            "institution": "Test institution",
            "name": "Test user",
        },
        "disorders": [],
        "features": [
            {"id": "HP:0001263", "label": "Global developmental delay", "observed": "yes"},
            {"id": "HP:0000252", "label": "Microcephaly", "observed": "yes"},
        ],
        "genomicFeatures": [
            {
                "gene": {"id": "3735"},
                "type": {"id": "SO:0001583", "label": "MISSENSE"},
                "variant": {
                    "alternateBases": "A",
                    "assembly": "GRCh37",
                    "end": 75665092,
                    "referenceBases": "G",
                    "referenceName": "16",
                    "start": 75665091,
                },
                "zygosity": 1,
            }
        ],
        "id": "P0001013",
        "label": "169_1-test",
    }
    return patient


@pytest.fixture(scope="function")
def match_objs():
    """Mock the results of an internal and an external match"""
    matches = [
        {  # External match where test_patient is the query and with results
            "_id": "match_1",
            "has_matches": True,
            "data": {
                "patient": {"id": "P0001058", "contact": {"href": "mailto:test_contact@email.com"}}
            },
            "results": [
                {
                    "node": {"id": "test_node1", "label": "Test Node 1"},
                    "patients": [{"patient": {"id": "patient1"}}, {"patient": {"id": "patient2"}}],
                },
                {
                    "node": {"id": "test_node2", "label": "Test Node 2"},
                    "patients": [{"patient": {"id": "patient3"}}],
                },
            ],
            "match_type": "external",
        },
        {  # Internal match where test_patient is the query and there are no results
            "_id": "match_2",
            "has_matches": False,
            "data": {
                "patient": {"id": "P0001058"},
                "contact": {"href": "mailto:test_contact@email.com"},
            },
            "results": [
                {
                    "node": {"id": "patientMatcher", "label": "patientMatcher server"},
                    "patients": [{"patient": {"id": "int_patient_id"}}],
                }
            ],
            "match_type": "internal",
        },
        {  #  Internal match where test_patient is among results
            "_id": "match_3",
            "has_matches": True,
            "data": {
                "patient": {
                    "id": "external_patient_1",
                    "contact": {"href": "mailto:test_contact@email.com"},
                }
            },
            "results": [
                {
                    "node": {"id": "test_node1", "label": "Test Node 1"},
                    "patients": [
                        {
                            "patient": {
                                "id": "P0001058",
                                "contact": {"href": "mailto:test_contact2@email.com"},
                            }
                        }
                    ],
                }
            ],
            "match_type": "internal",
        },
    ]
    return matches


@pytest.fixture()
def patient_37():
    """A patient with a variant in genome assembly GRCh38"""

    patient = {
        "patient": {
            "id": "patient_id",
            "contact": {"name": "Contact Name", "href": "mailto:contact_name@mail.com"},
            "features": [{"id": "HP:0009623"}],
            "genomicFeatures": [
                {
                    "gene": {"id": "GUCY2C"},
                    "variant": {
                        "assembly": "GRCh37",
                        "referenceName": "12",
                        "start": 14794075,
                        "end": 14794076,
                        "referenceBases": "C",
                        "alternateBases": "T",
                    },
                }
            ],
        }
    }
    return patient


@pytest.fixture(scope="function")
def gpx4_patients(json_patients):
    """Return all patients with variants in GPX4 gene"""

    patient_list = []

    for patient in json_patients:
        gpx4 = False
        if patient.get("genomicFeatures") is None:
            continue
        for g_feature in patient["genomicFeatures"]:
            if g_feature["gene"]["id"] == "GPX4":
                gpx4 = True
        if gpx4:
            patient_list.append(patient)
    return patient_list


@pytest.fixture(scope="function")
def json_patients(demo_data_path):
    """Returns a list of 50 MME test patients from demo data"""
    patients = {}
    with open(demo_data_path) as json_data:
        patients = json.load(json_data)
    return patients



# -*- coding: utf-8 -*-
import logging
import os

from patientMatcher.resources import path_to_hpo_terms
from patientMatcher.server import configure_email_error_logging, create_app
from patientMatcher.server.__init__ import available_phenotype_resources
from patientMatcher.utils.notify import TlsSMTPHandler


def test_env_vars_parsing(monkeypatch):
    """Test that the env vars are correctly parsed when creating the app"""
    # GIVEN some params provided as env vars
    monkeypatch.setenv("DEBUG", "True")
    monkeypatch.setenv("TESTING", "False")
    monkeypatch.setenv("DB_PORT", "27017")
    monkeypatch.setenv("MAX_GT_SCORE", "0.7")
    monkeypatch.setenv("MAX_PHENO_SCORE", "0.3")
    monkeypatch.setenv("MAX_RESULTS", "3")
    monkeypatch.setenv("SCORE_THRESHOLD", "0.5")
    monkeypatch.setenv("MAIL_PORT", "586")
    monkeypatch.setenv("NOTIFY_COMPLETE", "False")

    app = create_app()

    # THEN the app should be created with the expected config values
    app.config["DEBUG"] is True
    app.config["TESTING"] is False
    app.config["DB_PORT"] == 27017
    app.config["MAX_GT_SCORE"] == 0.7
    app.config["MAX_PHENO_SCORE"] == 0.3
    app.config["MAX_RESULTS"] == 3
    app.config["SCORE_THRESHOLD"] == 0.5
    app.config["MAIL_PORT"] == 586
    app.config["NOTIFY_COMPLETE"] is False


def test_create_app():
    """Tests the function that creates the app"""

    assert create_app()


def test_available_phenotype_resources():
    """Test function response when resource files (HPO terms and phenotype annotations) are both present"""
    assert available_phenotype_resources()


def test_available_phenotype_resources_missing_resource():
    """Test function response when one resource file is missing"""

    # GIVEN one of the required files missing (here it's been renamed)
    temp_file = ".".join([path_to_hpo_terms, "temp"])
    os.rename(path_to_hpo_terms, temp_file)

    # The function should return False
    assert available_phenotype_resources() is False

    # Revert original resource name
    os.rename(temp_file, path_to_hpo_terms)


def test_app_logging(monkeypatch):
    """Test the app error logging via email"""

    #

    # GIVEN an app with an ADMIN and configured email error logging params
    monkeypatch.setenv("DEBUG", "True")
    monkeypatch.setenv("ADMINS", '["admin@test.se"]')
    monkeypatch.setenv("MAIL_SERVER", "smtp.gmail.com")
    monkeypatch.setenv("MAIL_PORT", "587")
    monkeypatch.setenv("MAIL_USERNAME", "server_email")
    monkeypatch.setenv("MAIL_PASSWORD", "server_pw")
    monkeypatch.setenv("MAIL_USE_TLS", "True")

    # WHEN the app is created
    app = create_app()

    # An app logger that logs from DEBUG level should be created
    assert app.logger
    assert app.logger.level == logging.DEBUG

    # A logger handler of type TlsSMTPHandler should be created
    logging_handler = app.logger.handlers[0]
    assert isinstance(logging_handler, TlsSMTPHandler)
    # The handler logs from ERROR level
    assert logging_handler.level == logging.ERROR



# -*- coding: utf-8 -*-
import json

import responses
from patientMatcher.__version__ import __version__
from patientMatcher.match.handler import patient_matches
from patientMatcher.parse.patient import mme_patient
from patientMatcher.server.controllers import validate_response
from patientMatcher.utils.add import add_node, backend_add_patient, load_demo_patients


def test_heartbeat(mock_app, database, test_client):
    """Test sending a GET request to see if app has a heartbeat"""

    # send a get request without being authorized
    response = mock_app.test_client().get("heartbeat")
    assert response.status_code == 401

    # add an authorized client to database
    ok_token = test_client["auth_token"]
    add_node(mongo_db=mock_app.db, obj=test_client, is_client=True)

    # make sure that the request using its token is valid
    response = mock_app.test_client().get("heartbeat", headers=auth_headers(ok_token))
    assert response.status_code == 200

    # Make sure that all important info is returned
    data = json.loads(response.data)
    assert data["disclaimer"] == mock_app.config.get("DISCLAIMER")
    assert data["heartbeat"]["version"] == __version__
    assert data["heartbeat"]["production"] is False
    assert isinstance(data["heartbeat"]["accept"], list)
    assert len(data["heartbeat"]["accept"]) > 0


def test_add_patient_no_auth(mock_app, gpx4_patients):
    """Test sending a POST request to server to add a patient without valid token"""

    assert len(gpx4_patients) == 2  # patients with variants in this gene
    patient_data = gpx4_patients[1]

    # try to add a patient without being authorized
    response = mock_app.test_client().post(
        "patient/add", data=json.dumps(patient_data), headers=unauth_headers()
    )
    assert response.status_code == 401


def test_add_patient_malformed_patient(mock_app, test_client, gpx4_patients, test_node):
    """Test sending a POST request to server to add a patient with malformed patient json"""

    patient_data = gpx4_patients[1]

    # Given a node with authorized token
    ok_token = test_client["auth_token"]

    add_node(mongo_db=mock_app.db, obj=test_client, is_client=True)
    add_node(mongo_db=mock_app.db, obj=test_node, is_client=False)  #

    # send a malformed json object using a valid auth token
    malformed_json = "{'_id': 'patient_id' }"
    response = mock_app.test_client().post(
        "patient/add", data=malformed_json, headers=auth_headers(ok_token)
    )
    # and check that you get the correct error code from server(400)
    assert response.status_code == 400


def test_add_patient_malformed_data(mock_app, test_client, gpx4_patients, test_node):
    """Test sending a POST request to server to add a patient with malformed data"""

    patient_data = gpx4_patients[1]

    # Given a node with authorized token
    ok_token = test_client["auth_token"]

    add_node(mongo_db=mock_app.db, obj=test_client, is_client=True)
    add_node(
        mongo_db=mock_app.db, obj=test_node, is_client=False
    )  # add a test node, to perform external matching

    # add a patient not conforming to MME API using a valid auth token
    response = mock_app.test_client().post(
        "patient/add", data=json.dumps(patient_data), headers=auth_headers(ok_token)
    )
    # and check that the server returns an error 422 (unprocessable entity)
    assert response.status_code == 422


@responses.activate
def test_add_patient(mock_app, test_client, gpx4_patients, test_node, database):
    """Test adding a patient by sending a POST request to the add endpoint with valid data"""

    # GIVEN a mocked Ensembl REST API converting gene symbol to Ensembl ID
    responses.add(
        responses.GET,
        f"https://grch37.rest.ensembl.org/xrefs/symbol/homo_sapiens/GPX4?external_db=HGNC",
        json=[{"id": "ENSG00000167468", "type": "gene"}],
        status=200,
    )

    patient_data = gpx4_patients[1]

    # Given a node with authorized token
    ok_token = test_client["auth_token"]

    add_node(mongo_db=mock_app.db, obj=test_client, is_client=True)
    add_node(
        mongo_db=mock_app.db, obj=test_node, is_client=False
    )  # add a test node, to perform external matching

    # a matches collection without documents
    assert database["matches"].find_one() is None

    # and an empty patients collection
    assert database["patients"].find_one() is None

    patient_obj = {"patient": patient_data}  # this is a valid patient object
    # WHEN the patient is added using the add enpoint
    response = mock_app.test_client().post(
        "patient/add", data=json.dumps(patient_obj), headers=auth_headers(ok_token)
    )
    assert response.status_code == 200

    # make sure that the POST request to add the patient triggers the matching request to an external node
    assert database["matches"].find_one()

    # There should be one patient in database now
    assert database["patients"].find_one()

    # the patient in database has label "Patient number 1"
    result = database["patients"].find_one({"label": "350_2-test"})
    # make sure patient's gene was converted to ensembl
    assert result["genomicFeatures"][0]["gene"]["id"].startswith("ENSG")
    # and that non-standard but informative _geneName field was added to genomic feature
    assert result["genomicFeatures"][0]["gene"]["_geneName"] == "GPX4"


def test_update_patient(mock_app, test_client, gpx4_patients, test_node, database):
    """Test updating a patient by sending a POST request to the add endpoint with valid data"""

    patient_data = gpx4_patients[1]
    # Modify patient's contact href with a simple email
    patient_data["contact"]["href"] = "somebody@test.se"

    # Given a node with authorized token
    ok_token = test_client["auth_token"]

    add_node(mongo_db=mock_app.db, obj=test_client, is_client=True)
    add_node(
        mongo_db=mock_app.db, obj=test_node, is_client=False
    )  # add a test node, to perform external matching

    # a matches collection without documents
    assert database["matches"].find_one() is None

    # and an empty patients collection
    assert database["patients"].find_one() is None

    # GIVEN a patient added using the add enpoint
    patient_obj = {"patient": patient_data}  # this is a valid patient object
    response = mock_app.test_client().post(
        "patient/add", data=json.dumps(patient_obj), headers=auth_headers(ok_token)
    )
    assert response.status_code == 200

    # WHEN the patient is updated using the same add endpoint
    patient_data["label"] = "modified patient label"
    patient_obj = {"patient": patient_data}  # this
    response = mock_app.test_client().post(
        "patient/add", data=json.dumps(patient_obj), headers=auth_headers(ok_token)
    )
    assert response.status_code == 200

    # Then there should still be one patient in the database
    updated_patient = database["patients"].find_one()
    # With an updated and valid email address
    assert updated_patient["contact"]["href"] == "mailto:somebody@test.se"

    # And the update has triggered an additional external patient matching
    results = database["matches"].find()
    assert len(list(results)) == 2


def test_metrics(mock_app, database, test_client, demo_data_path, match_objs):
    """Testing viewing the list of patients on server for authorized users"""

    # load demo data of 50 test patients
    inserted_ids = load_demo_patients(demo_data_path, database)
    assert len(inserted_ids) == 50  # 50 test cases should be loaded

    # load mock matches into database
    database.matches.insert_many(match_objs)
    results = database.matches.find()
    assert len(list(results)) == 3

    # send a get request without being authorized
    response = mock_app.test_client().get("metrics")
    assert response.status_code == 200

    data = json.loads(response.data)
    assert data["disclaimer"]  # disclaimer should be returned
    metrics = data["metrics"]

    assert metrics["numberOfCases"] == 50
    assert metrics["numberOfSubmitters"] > 0
    assert metrics["numberOfGenes"] > metrics["numberOfUniqueGenes"]
    assert metrics["numberOfVariants"] > metrics["numberOfUniqueVariants"]
    assert metrics["numberOfFeatures"] > metrics["numberOfUniqueFeatures"]
    assert metrics["numberOfCasesWithDiagnosis"] > 0
    assert metrics["numberOfUniqueGenesMatched"] == 0  # no gene was provided in match_obj results
    assert metrics["numberOfRequestsReceived"] == 2  # Sent 2 requests
    assert metrics["numberOfPotentialMatchesSent"] == 1  # Just one has returned results


def test_nodes_view(mock_app, database, test_node, test_client):
    """testing viewing the list of connected nodes as an authenticated client"""

    # insert a test client in database
    add_node(mongo_db=database, obj=test_client, is_client=True)
    ok_token = test_client["auth_token"]

    # If you try to see nodes without being authorized
    response = mock_app.test_client().get("nodes")
    # You should get a not authorized code from server
    assert response.status_code == 401

    # since there are no connected nodes in database
    assert database["nodes"].find_one() is None
    # When you send an authorized request
    response = mock_app.test_client().get("nodes", headers=auth_headers(ok_token))
    data = json.loads(response.data)
    # you shoud get an empty list
    assert data == []

    # insert a test node in database
    add_node(mongo_db=database, obj=test_node, is_client=False)
    # When you send an authorized request
    response = mock_app.test_client().get("nodes", headers=auth_headers(ok_token))
    data = json.loads(response.data)
    # this time you should get a list with one element
    assert len(data) == 1
    # and the id of the element is the id of the node
    assert data[0]["id"] == test_node["_id"]


def test_delete_patient(mock_app, database, gpx4_patients, test_client, match_objs):
    """Test deleting a patient from database by sending a DELETE request"""

    # load 2 patients from demo data in mock database
    assert len(gpx4_patients) == 2
    inserted_ids = []
    for pat in gpx4_patients:
        # convert patient in mme patient type (convert also gene to ensembl)
        mme_pat = mme_patient(pat, True)
        inserted_ids.append(backend_add_patient(database, mme_pat))

    assert len(inserted_ids) == 2

    # 50 cases present on patients collection
    delete_id = "P0001058"

    # try to delete patient without auth token:
    response = mock_app.test_client().delete("".join(["patient/delete/", delete_id]))
    assert response.status_code == 401

    # Add a valid client node
    ok_token = test_client["auth_token"]
    add_node(mongo_db=mock_app.db, obj=test_client, is_client=True)

    # Send delete request providing a valid token but a non valid id
    response = mock_app.test_client().delete(
        "".join(["patient/delete/", "not_a_valid_ID"]), headers=auth_headers(ok_token)
    )
    assert response.status_code == 200
    data = json.loads(response.data)
    # but server returns error
    assert (
        data["message"] == "ERROR. Could not delete a patient with ID not_a_valid_ID from database"
    )

    assert database["matches"].find_one() is None  # no matches in database
    # insert into database some mock matching objects
    database["matches"].insert_many(match_objs)

    # patient "delete_id" should have two associated matches in database
    results = database["matches"].find({"data.patient.id": delete_id})
    assert len(list(results)) == 2

    # Send valid patient ID and valid token
    response = mock_app.test_client().delete(
        "".join(["patient/delete/", delete_id]), headers=auth_headers(ok_token)
    )
    assert response.status_code == 200

    # make sure that the patient was removed from database
    results = database["patients"].find()
    assert len(list(results)) == 1

    # make sure that patient matches are also gone
    results = database["matches"].find()
    assert len(list(results)) == 1


def test_patient_matches(mock_app, database, match_objs, test_client):
    """testing the endpoint that retrieves the matchings by patient ID"""

    # Add a valid client node
    ok_token = test_client["auth_token"]
    add_node(mongo_db=mock_app.db, obj=test_client, is_client=True)

    # start from a database with no matches
    assert database["matches"].find_one() is None
    # import mock matches into datababase
    database["matches"].insert_many(match_objs)
    # database now should have two matching objects

    # test endpoint to get matches by ID
    # test by sending a non-authorized request
    response = mock_app.test_client().get("matches/P0001058")
    # response gives a 401 code (not authorized)
    assert response.status_code == 401

    # try with an authorized request with a used ID that is not in database
    response = mock_app.test_client().get("matches/unknown_patient", headers=auth_headers(ok_token))
    # response gives success
    assert response.status_code == 200
    data = json.loads(response.data)
    # but the patient is not found by server
    assert (
        data["message"] == "Could not find any matches in database for patient ID unknown_patient"
    )

    # Try with authenticates request and valid patient
    response = mock_app.test_client().get("matches/P0001058", headers=auth_headers(ok_token))
    # response gives success
    assert response.status_code == 200
    data = json.loads(response.data)
    # and there are matches in it
    assert (
        len(data["matches"]) == 2
    )  # 2 matches returned because endpoint returns only matches with results

    # Test that there are actually 3 matches by calling directly the function returning matches
    matches = patient_matches(
        database=database, patient_id="P0001058", type=None, with_results=False
    )
    assert len(matches) == 3
    for match in matches:
        for result in match["results"]:
            for patient in result["patients"]:
                assert patient["patient"]["id"]

    # Call the same function to get only external matches
    matches = patient_matches(
        database=database, patient_id="P0001058", type="external", with_results=False
    )
    assert len(matches) == 1

    # Call the same function to get only external matches
    matches = patient_matches(
        database=database, patient_id="P0001058", type="internal", with_results=False
    )
    assert len(matches) == 2


@responses.activate
def test_match_hgnc_symbol_patient(mock_app, gpx4_patients, test_client, database):
    """Testing matching patient with gene symbol against patientMatcher database (internal matching)"""

    # GIVEN a mocked Ensembl REST API converting gene symbol to Ensembl ID
    responses.add(
        responses.GET,
        f"https://grch37.rest.ensembl.org/xrefs/symbol/homo_sapiens/GPX4?external_db=HGNC",
        json=[{"id": "ENSG00000167468", "type": "gene"}],
        status=200,
    )

    # GIVEN a mocked Ensembl gene lookup service:
    responses.add(
        responses.GET,
        f"https://grch37.rest.ensembl.org/lookup/id/ENSG00000167468",
        json=[{"display_name": "GPX4"}],
        status=200,
    )

    # GIVEN a mocked liftover service:
    responses.add(
        responses.GET,
        f"https://grch37.rest.ensembl.org/map/human/GRCh37/19:1105813..1105814/GRCh38?content-type=application/json",
        json=[],
        status=200,
    )
    responses.add(
        responses.GET,
        f"https://grch37.rest.ensembl.org/map/human/GRCh37/19:1106232..1106238/GRCh38?content-type=application/json",
        json=[],
        status=200,
    )

    # add an authorized client to database
    ok_token = test_client["auth_token"]
    add_node(mongo_db=mock_app.db, obj=test_client, is_client=True)

    query_patient = {"patient": gpx4_patients[0]}
    assert query_patient["patient"]["genomicFeatures"][0]["gene"]["id"] == "GPX4"

    # load 2 test patient in mock database
    assert len(gpx4_patients) == 2
    inserted_ids = []
    for pat in gpx4_patients:
        # convert patient in mme patient type (convert also gene to ensembl)
        mme_pat = mme_patient(pat, True)
        inserted_ids.append(backend_add_patient(database, mme_pat))

    assert len(inserted_ids) == 2

    # test the API response validator with non valid patient data:
    malformed_match_results = {"results": "fakey_results"}
    assert validate_response(malformed_match_results) == 422

    # make sure that there are no patient matches in the 'matches collection'
    assert database["matches"].find_one() is None

    # send a POST request to match patient with patients in database
    response = mock_app.test_client().post(
        "/match", data=json.dumps(query_patient), headers=auth_headers(ok_token)
    )
    assert response.status_code == 200  # POST request should be successful
    data = json.loads(response.data)
    # data should contain results and the max number of results is as defined in the config file
    assert len(data["results"]) == 2
    assert type(data["results"]) == list  # which is a list
    assert "patient" in data["results"][0]  # of patients
    assert "score" in data["results"][0]  # with matching scores
    assert "contact" in data["results"][0]["patient"]  # contact info should be available as well

    # make sure that there are match object is created in db for this internal matching
    match = database["matches"].find_one()
    for res in match["results"]:
        for pat in res["patients"]:
            assert pat["patient"]["contact"]  # each result should have a contact person
            assert pat["score"]["patient"] > 0


@responses.activate
def test_match_ensembl_patient(mock_app, test_client, gpx4_patients, database):
    """Test matching patient with ensembl gene against patientMatcher database (internal matching)"""

    # GIVEN a mocked Ensembl REST API converting gene symbol to Ensembl ID
    responses.add(
        responses.GET,
        f"https://grch37.rest.ensembl.org/xrefs/symbol/homo_sapiens/GPX4?external_db=HGNC",
        json=[{"id": "ENSG00000167468", "type": "gene"}],
        status=200,
    )

    # GIVEN a mocked Ensembl gene lookup service:
    responses.add(
        responses.GET,
        f"https://grch37.rest.ensembl.org/lookup/id/ENSG00000167468",
        json=[{"display_name": "GPX4"}],
        status=200,
    )

    # GIVEN a mocked liftover service:
    responses.add(
        responses.GET,
        f"https://grch37.rest.ensembl.org/map/human/GRCh37/19:1105813..1105814/GRCh38?content-type=application/json",
        json=[],
        status=200,
    )
    responses.add(
        responses.GET,
        f"https://grch37.rest.ensembl.org/map/human/GRCh37/19:1106232..1106238/GRCh38?content-type=application/json",
        json=[],
        status=200,
    )

    # add an authorized client to database
    ok_token = test_client["auth_token"]
    add_node(mongo_db=mock_app.db, obj=test_client, is_client=True)

    query_patient = {"patient": mme_patient(gpx4_patients[0], True)}
    assert query_patient["patient"]["genomicFeatures"][0]["gene"]["id"].startswith("ENSG")

    # load 2 test patient in mock database
    assert len(gpx4_patients) == 2
    inserted_ids = []
    for pat in gpx4_patients:
        # convert patient in mme patient type (convert also gene to ensembl)
        mme_pat = mme_patient(pat, True)
        inserted_ids.append(backend_add_patient(database, mme_pat))

    assert len(inserted_ids) == 2

    # make sure that there are no patient matches in the 'matches collection'
    assert database["matches"].find_one() is None

    # send a POST request to match patient with patients in database
    response = mock_app.test_client().post(
        "/match", data=json.dumps(query_patient), headers=auth_headers(ok_token)
    )
    assert response.status_code == 200  # POST request should be successful
    data = json.loads(response.data)
    # data should contain results and the max number of results is as defined in the config file
    assert len(data["results"]) == 2
    assert type(data["results"]) == list  # which is a list
    assert "patient" in data["results"][0]  # of patients
    assert "score" in data["results"][0]  # with matching scores
    assert "contact" in data["results"][0]["patient"]  # contact info should be available as well

    # make sure that a match object is created in db for this internal matching
    match = database["matches"].find_one()
    for res in match["results"]:
        for pat in res["patients"]:
            assert pat["patient"]["contact"]  # each result should have a contact person
            assert pat["score"]["patient"] > 0
    # and query patient should have hgnc gene symbol saved as non-standard _geneName field
    assert match["data"]["patient"]["genomicFeatures"][0]["gene"]["_geneName"] == "GPX4"


@responses.activate
def test_match_entrez_patient(mock_app, test_client, gpx4_patients, database):
    """Test matching patient with ensembl gene against patientMatcher database (internal matching)"""

    # GIVEN a mocked Ensembl REST API converting gene symbol to Ensembl ID
    responses.add(
        responses.GET,
        f"https://grch37.rest.ensembl.org/xrefs/symbol/homo_sapiens/GPX4?external_db=HGNC",
        json=[{"id": "ENSG00000167468", "type": "gene"}],
        status=200,
    )

    # GIVEN a mocked Ensembl gene lookup service:
    responses.add(
        responses.GET,
        f"https://grch37.rest.ensembl.org/lookup/id/ENSG00000167468",
        json=[{"display_name": "GPX4"}],
        status=200,
    )

    # GIVEN a mocked liftover service:
    responses.add(
        responses.GET,
        f"https://grch37.rest.ensembl.org/map/human/GRCh37/19:1105813..1105814/GRCh38?content-type=application/json",
        json=[],
        status=200,
    )
    responses.add(
        responses.GET,
        f"https://grch37.rest.ensembl.org/map/human/GRCh37/19:1106232..1106238/GRCh38?content-type=application/json",
        json=[],
        status=200,
    )

    # add an authorized client to database
    ok_token = test_client["auth_token"]
    add_node(mongo_db=mock_app.db, obj=test_client, is_client=True)

    query_patient = gpx4_patients[0]
    query_patient = {"patient": gpx4_patients[0]}
    for feat in query_patient["patient"]["genomicFeatures"]:
        feat["gene"]["id"] == "2879"  # entrez id for GPX4

    # load 2 test patient in mock database
    assert len(gpx4_patients) == 2
    inserted_ids = []
    for pat in gpx4_patients:
        # convert patient in mme patient type (convert also gene to ensembl)
        mme_pat = mme_patient(pat, True)
        inserted_ids.append(backend_add_patient(database, mme_pat))

    assert len(inserted_ids) == 2

    # make sure that there are no patient matches in the 'matches collection'
    assert database["matches"].find_one() is None

    # send a POST request to match patient with patients in database
    response = mock_app.test_client().post(
        "/match", data=json.dumps(query_patient), headers=auth_headers(ok_token)
    )
    assert response.status_code == 200  # POST request should be successful
    data = json.loads(response.data)
    # data should contain results and the max number of results is as defined in the config file
    assert len(data["results"]) == 2
    assert type(data["results"]) == list  # which is a list
    assert "patient" in data["results"][0]  # of patients
    assert "score" in data["results"][0]  # with matching scores
    assert "contact" in data["results"][0]["patient"]  # contact info should be available as well

    # make sure that a match object is created in db for this internal matching
    match = database["matches"].find_one()
    for res in match["results"]:
        for pat in res["patients"]:
            assert pat["patient"]["contact"]  # each result should have a contact person
            assert pat["score"]["patient"] > 0
    # and query patient should have hgnc gene symbol saved as non-standard _geneName field
    assert match["data"]["patient"]["genomicFeatures"][0]["gene"]["_geneName"] == "GPX4"


def test_match_external(mock_app, test_client, test_node, database, json_patients):
    """Testing the view that is sending post request to trigger matches on external nodes"""

    # add an authorized client to database
    ok_token = test_client["auth_token"]
    add_node(
        mongo_db=mock_app.db, obj=test_client, is_client=True
    )  # required to trigger external matches

    a_patient = json_patients[0]
    parsed_patient = mme_patient(a_patient)

    # insert patient into mock database:
    assert database["patients"].find_one() is None
    inserted_id = database["patients"].insert_one(parsed_patient).inserted_id
    assert database["patients"].find_one()

    # send an un-authorized match request to server
    response = mock_app.test_client().post("".join(["/match/external/", inserted_id]))
    # server should return 401 (not authorized)
    assert response.status_code == 401

    # send an authorized request with a patient ID that doesn't exist on server:
    response = mock_app.test_client().post(
        "".join(["/match/external/", "not_a_valid_ID"]), headers=auth_headers(ok_token)
    )
    # Response is valid
    assert response.status_code == 200
    data = json.loads(response.data)
    # but server returns error
    assert data["message"] == "ERROR. Could not find any patient with ID not_a_valid_ID in database"

    # there are no matches in mock database
    assert database["matches"].find_one() is None
    # after sending an authorized request with a patient ID that exists on database

    # Check that external matching doesn't work if there are no connected nodes:
    response = mock_app.test_client().post(
        "".join(["/match/external/", inserted_id]), headers=auth_headers(ok_token)
    )
    assert response.status_code == 200
    data = json.loads(response.data)
    assert data["message"] == "Could not find any other node connected to this MatchMaker server"

    # Try to send a request for a match on a node that does not exist
    response = mock_app.test_client().post(
        "".join(["/match/external/", inserted_id, "?node=meh"]), headers=auth_headers(ok_token)
    )
    assert response.status_code == 200
    data = json.loads(response.data)
    # And check that node not found is in response message
    assert data["message"] == "ERROR. Could not find any connected node with id meh in database"

    # insert a connected node
    add_node(mongo_db=mock_app.db, obj=test_node, is_client=False)  # required for external matches
    # send a request to match patients against all nodes
    response = mock_app.test_client().post(
        "".join(["/match/external/", inserted_id]), headers=auth_headers(ok_token)
    )

    # Response should be valid
    assert response.status_code == 200
    # And a new match should be created in matches collection
    assert database["matches"].find_one()

    # send a request to match patients against the specific existing node:
    response = mock_app.test_client().post(
        "".join(["/match/external/", inserted_id, "?node=", test_node["_id"]]),
        headers=auth_headers(ok_token),
    )
    # Response should be valid
    assert response.status_code == 200

    # And a new match should be created in matches collection. So total matches are 2
    results = database["matches"].find()
    assert len(list(results)) == 2


def unauth_headers():
    head = {
        "Content-Type": "application/vnd.ga4gh.matchmaker.v1.0+json",
        "Accept": ["application/vnd.ga4gh.matchmaker.v1.0+json", "application/json"],
        "X-Auth-Token": "wrong_token",
    }
    return head


def auth_headers(test_token):
    head = {
        "Content-Type": "application/vnd.ga4gh.matchmaker.v1.0+json",
        "Accept": ["application/vnd.ga4gh.matchmaker.v1.0+json", "application/json"],
        "X-Auth-Token": test_token,
    }
    return head



# -*- coding: utf-8 -*-
import json
import logging
from logging.handlers import SMTPHandler

from flask_mail import Message
from patientMatcher.__version__ import __version__

LOG = logging.getLogger(__name__)


class TlsSMTPHandler(SMTPHandler):
    """Class for handling error logging system"""

    def emit(self, record):
        """Emit a record.
        Format the record and send it to the specified admin addressees.
        """
        try:
            import smtplib

            try:
                from email.utils import formatdate
            except ImportError:
                formatdate = self.date_time
            port = self.mailport
            if not port:
                port = smtplib.SMTP_PORT
            smtp = smtplib.SMTP(self.mailhost, port)
            msg = self.format(record)
            msg = "From: %s\r\nTo: %s\r\nSubject: %s\r\nDate: %s\r\n\r\n%s" % (
                self.fromaddr,
                ",".join(self.toaddrs),
                self.getSubject(record),
                formatdate(),
                msg,
            )
            if self.username:
                smtp.ehlo()  # For 'tls', add this line
                smtp.starttls()  # For 'tls', add this line
                smtp.ehlo()  # For 'tls', add this line
                smtp.login(self.username, self.password)
            smtp.sendmail(self.fromaddr, self.toaddrs, msg)
            smtp.quit()

        except Exception as ex:
            LOG.warning(ex)


def notify_match_internal(database, match_obj, admin_email, mail, notify_complete):
    """Send an email to patient contacts after an internal match

    Args:
        database(pymongo.database.Database): patientMatcher database
        match_obj(dict): an object containing both query patient(dict) and matching results(list)
        admin_email(str): email of the server admin
        mail(flask_mail.Mail): an email instance
        notify_complete(bool): set to False to NOT notify variants and phenotype terms by email
    """
    # Internal matching can be triggered by a patient in the same database or by a patient on a connected node.
    # In the first case notify both querier contact and contacts in the result patients.
    # in the second case notify only contacts from patients in the results list.
    sender = admin_email
    patient_id = None
    patient_label = None
    results = None
    recipient = None
    email_subject = "MatchMaker Exchange: new patient match available."
    email_body = None

    # check if query patient already belongs to patientMatcher database:
    internal_patient = database["patients"].find_one({"_id": match_obj["data"]["patient"]["id"]})
    if internal_patient:
        # If patient used for the search is in patientMatcher database, notify querier as well:
        patient_id = match_obj["data"]["patient"]["id"]
        patient_label = match_obj["data"]["patient"].get("label")
        recipient = match_obj["data"]["patient"]["contact"]["href"][7:]
        email_body = active_match_email_body(
            patient_id=patient_id,
            match_results=match_obj["results"],
            patient_label=patient_label,
            external_match=False,
            notify_complete=notify_complete,
        )
        LOG.info(
            "Sending an internal match notification for query patient with ID:{0}. Patient contact: {1}".format(
                patient_id, recipient
            )
        )

        kwargs = dict(subject=email_subject, html=email_body, sender=sender, recipients=[recipient])
        message = Message(**kwargs)
        # send email using flask_mail
        try:
            mail.send(message)
        except Exception as err:
            LOG.error(
                "An error occurred while sending an internal match notification: {}".format(err)
            )

    # Loop over the result patients and notify their contact about the matching with query patient
    for result in match_obj["results"][0][
        "patients"
    ]:  # this list has only one element since there is only one internal node

        patient_id = result["patient"]["id"]

        # do not notify when patient in results is the same as the one used for query
        if internal_patient and internal_patient["_id"] == patient_id:
            continue

        patient_label = result["patient"].get("label")
        recipient = result["patient"]["contact"]["href"][7:]
        email_body = passive_match_email_body(
            patient_id, match_obj["data"]["patient"], patient_label, notify_complete
        )
        LOG.info(
            "Sending an internal match notification for match result with ID {}".format(patient_id)
        )

        kwargs = dict(subject=email_subject, html=email_body, sender=sender, recipients=[recipient])
        message = Message(**kwargs)
        # send email using flask_mail
        try:
            mail.send(message)
        except Exception as err:
            LOG.error(
                "An error occurred while sending an internal match notification: {}".format(err)
            )


def notify_match_external(match_obj, admin_email, mail, notify_complete):
    """Send an email to patients contacts to notify a match on external nodes

    Args:
        match_obj(dict): an object containing both query patient(dict) and matching results(list)
        admin_email(str): email of the server admin
        mail(flask_mail.Mail): an email instance
        notify_complete(bool): set to False to NOT notify variants and phenotype terms by email
    """
    sender = admin_email
    patient_id = match_obj["data"]["patient"]["id"]
    patient_label = match_obj["data"]["patient"].get("label")
    recipient = match_obj["data"]["patient"]["contact"]["href"][7:]
    email_subject = "MatchMaker Exchange: new patient match available."
    email_body = active_match_email_body(
        patient_id=patient_id,
        match_results=match_obj["results"],
        patient_label=patient_label,
        external_match=True,
        notify_complete=notify_complete,
    )
    LOG.info(
        "Sending an external match notification for query patient with ID {0}. Patient contact: {1}".format(
            patient_id, recipient
        )
    )

    kwargs = dict(subject=email_subject, html=email_body, sender=sender, recipients=[recipient])
    message = Message(**kwargs)
    # send email using flask_mail
    try:
        mail.send(message)
    except Exception as err:
        LOG.error("An error occurred while sending an external match notification: {}".format(err))


def active_match_email_body(
    patient_id, match_results, patient_label=None, external_match=False, notify_complete=False
):
    """Returns the body message of the notification email when the patient was used as query patient

    Args:
        patient_id(str): the ID of the patient submitted by the  MME user which will be notified
        match_results(list): a list of patients which match with the patient whose contact is going to be notified
        external_match(bool): True == match in connected nodes, False == match with other patients in database
        patient_label(str): the label of the patient submitted by the  MME user which will be notified (not mandatory field)
        notify_complete(bool): set to False to NOT notify variants and phenotype terms by email

    Returns:
        html(str): the body message
    """
    search_type = "against the internal database of MatchMaker patients"
    if external_match:
        search_type = "against external nodes connected to MatchMaker"

    html = """
        ***This is an automated message, please do not reply to this email.***<br><br>
        <strong>MatchMaker Exchange patient matching notification:</strong><br><br>
        Patient with ID <strong>{0}</strong>, label <strong>{1}</strong>.
        PatientMatcher v.{2}.<br>
        This search returned these potential matches</strong>:<br>
        <strong>{3}</strong><br>
        You might directly contact the matching part using the address specified in patient's data or review matching
        results in the portal you used to submit your patient.
        <br><br>
        Kind regards,<br>
        The PatientMatcher team
    """.format(
        patient_id, patient_label, __version__, html_format(match_results, 0, notify_complete)
    )

    return html


def passive_match_email_body(
    patient_id, matched_patient, patient_label=None, notify_complete=False
):
    """Returns the body message of the notification email when the patient was used as query patient

    Args:
        patient_id(str): the ID of the patient submitted by the MME user which will be notified
        matched_patient(dict): a patient object
        patient_label(str): the label of the patient submitted by the  MME user which will be notified (not mandatory field)
        notify_complete(bool): set to False to NOT notify variants and phenotype terms by email

    Returns:
        html(str): the body message
    """
    html = """
        ***This is an automated message, please do not reply.***<br>
        <strong>MatchMaker Exchange patient matching notification:</strong><br><br>
        PatientMatcher v.{0}.<br>
        Patient with <strong>ID {1}</strong>,<strong> label {2}</strong> was recently returned as a match result
        in a search performed using a patient with these specifications:<br>
        <strong>{3}</strong><br>
        You might directly contact the matching part using the address specified in patient's data or review matching
        results in the portal you used to submit your patient.
        <br><br>
        Kind regards,<br>
        The PatientMatcher team
    """.format(
        __version__, patient_id, patient_label, html_format(matched_patient, 0, notify_complete)
    )

    return html


def admins_email_format(admins_emails):
    """Formats the ADMINS value provided by a server admin via env vars"""
    if isinstance(admins_emails, str):
        admins_emails = json.loads(admins_emails)
    return admins_emails


def html_format(obj, indent=0, notify_complete=False):
    """Formats one or more patient objects to a nice html string

    Args:
        obj(list): a list of patient objects or a patient object
        notify_complete(bool): set to False to NOT notify variants and phenotype terms by email
    """
    if isinstance(obj, list):  # a list pf match results
        htmls = []
        for k in obj:
            htmls.append(html_format(obj=k, indent=indent + 1, notify_complete=notify_complete))

        return '[<div style="margin-left: %dem">%s</div>]' % (indent, ",<br>".join(htmls))

    if isinstance(obj, dict):  # patient object
        htmls = []
        for k, v in obj.items():
            if notify_complete or k in [
                "node",
                "patients",
                "patient",
                "contact",
                "id",
                "name",
                "href",
                "email",
                "institution",
            ]:
                htmls.append(
                    "<span style='font-style: italic; color: #888'>%s</span>: %s"
                    % (k, html_format(obj=v, indent=indent + 1, notify_complete=notify_complete))
                )

        return '{<div style="margin-left: %dem">%s</div>}' % (indent, ",<br>".join(htmls))

    return str(obj)



# -*- coding: utf-8 -*-
import logging
from datetime import date

LOG = logging.getLogger(__name__)


def general_metrics(db):
    """Create an object with database metrics

    Args:
        db(pymongo.database.Database)

    Returns:
        metrics(dict): According to the MME API it should be a dictionary like this:
            {
                "metrics": {
                    "numberOfCases": 0,
                    "numberOfSubmitters": 0,
                    "numberOfGenes": 0,
                    "numberOfUniqueGenes": 0,
                    "numberOfVariants": 0,
                    "numberOfUniqueVariants": 0,
                    "numberOfFeatures": 0,
                    "numberOfUniqueFeatures": 0,
                    "numberOfFeatureSets": 0, # endpoint is not returning this, at the moment
                    "numberOfUniqueGenesMatched": 0,
                    "numberOfCasesWithDiagnosis": 0,
                    "numberOfRequestsReceived": 0,
                    "numberOfPotentialMatchesSent": 0,
                    "dateGenerated": "2017-08-24",

                },
                "disclaimer": "Disclaimer text...",
                "terms": "Terms text..."
            }
    """
    # get gene/occurrence for all genes in db
    n_genes = 0
    gene_occurrs = item_occurrence(
        db, "genomicFeatures", "genomicFeatures.gene", "genomicFeatures.gene.id"
    )
    for gene_count in gene_occurrs:
        n_genes += gene_count["count"]

    # get numberOfUniqueVariants/occurrence for all variants in db
    variant_occurr = item_occurrence(
        db, "genomicFeatures", "genomicFeatures.variant", "genomicFeatures.variant"
    )
    n_vars = 0
    for var in variant_occurr:
        n_vars += var.get("count")

    # get feature/occurrence for all features in db
    n_feat = 0
    feat_occurr = item_occurrence(db, "features", "features.id")
    for feat in feat_occurr:
        n_feat += feat.get("count")

    # include in unique_gene_matches only matches actively returned by the server (internal)
    match_type = {"match_type": "internal"}
    unique_gene_matches = db.matches.distinct(
        "results.patients.patient.genomicFeatures.gene", match_type
    )

    n_cases = sum(1 for i in db.patients.find())
    n_cases_diagnosis = sum(
        1 for i in db.patients.find({"disorders": {"$exists": True, "$ne": []}})
    )
    n_requests = sum(1 for i in db.matches.find({"match_type": "internal"}))
    n_positive_matches = sum(
        1 for i in db.matches.find({"match_type": "internal", "has_matches": True})
    )
    metrics = {
        "numberOfCases": n_cases,
        "numberOfSubmitters": len(db.patients.distinct("contact.href")),
        "numberOfGenes": n_genes,
        "numberOfUniqueGenes": len(db.patients.distinct("genomicFeatures.gene")),
        "numberOfVariants": n_vars,
        "numberOfUniqueVariants": len(db.patients.distinct("genomicFeatures.variant")),
        "numberOfFeatures": n_feat,
        "numberOfUniqueFeatures": len(db.patients.distinct("features.id")),
        "numberOfUniqueGenesMatched": len(unique_gene_matches),
        "numberOfCasesWithDiagnosis": n_cases_diagnosis,
        "numberOfRequestsReceived": n_requests,
        "numberOfPotentialMatchesSent": n_positive_matches,
        "dateGenerated": str(date.today()),
    }
    return metrics


def item_occurrence(db, unw1, group, unw2=None):
    """Get a list of item/occurrence in patient collection

    Args:
        db(pymongo.database.Database)
        unw1(string): first nested unwind item
        group(string): item to group results by
        unw2(string): second nested unwind item # none if nested level is missing

    Returns:
        item_occurr(list) example: [{'id':'item_obj', 'count': item_occurrence}, ..]
    """
    # create query pipeline
    pipeline = [{"$unwind": "".join(["$", unw1])}]
    if unw2:
        pipeline.append({"$unwind": "".join(["$", unw2])})
    pipeline.append({"$group": {"_id": "".join(["$", group]), "count": {"$sum": 1}}})
    item_occurr = list(db.patients.aggregate(pipeline))
    return item_occurr



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import logging
import os

from patientMatcher.parse.patient import disorders_to_omim, features_to_hpo
from patientMatcher.resources import path_to_hpo_terms, path_to_phenotype_annotations
from patientMatcher.server.extensions import hpo as hpo_extension
from patientMatcher.server.extensions import hpoic
from patientMatcher.utils.patient import Patient, pheno_similarity_score_simgic

LOG = logging.getLogger(__name__)


def match(database, max_score, features, disorders):
    """Handles phenotype matching algorithm

    Args:
        database(pymongo.database.Database)
        max_score(float): a number between 0 and 1
        features(list): a list of phenotype feature objects (example ID = HP:0008619)
        disorders(list): a list of OMIM diagnoses (example ID = MIM:616007 )

    Returns:
        matches(dict): a dictionary of patient matches with phenotype matching score
    """
    matches = {}

    hpo_terms = []
    omim_terms = []
    query_fields = []

    if features:  # at least one HPO term is specified
        hpo_terms = features_to_hpo(features)
        # compare against all cases which also have features (HPO terms)
        query_fields.append({"features": {"$exists": True, "$ne": []}})

    if disorders:  # at least one OMIM term was provided
        omim_terms = disorders_to_omim(disorders)
        query_fields.append({"disorders.id": {"$in": omim_terms}})

    # build a database query taking into account patient features (HPO terms) and disorders (omim)
    if len(query_fields) > 0:
        query = {"$or": query_fields}
        pheno_matching_patients = list(database["patients"].find(query))
        LOG.info(
            "\n\nFOUND {} patients matching query: {}\n\n".format(
                len(pheno_matching_patients), query
            )
        )

        for i in range(len(pheno_matching_patients)):
            patient = pheno_matching_patients[i]
            similarity = evaluate_pheno_similariy(
                hpoic, hpo_extension, hpo_terms, omim_terms, patient, max_score
            )

            match = {
                "patient_obj": patient,
                "pheno_score": similarity,
            }
            matches[patient["_id"]] = match

    return matches


def evaluate_pheno_similariy(
    hpoic, hpo, hpo_terms, disorders, pheno_matching_patient, max_similarity
):
    """Evaluates the similarity of two patients based on phenotype features

    Args:
        hpoic(class) : the information content for the HPO
        hpo(class): an instance of the class for interacting with HPO
        hpo_terms(list): HPO terms of the query patient
        disorders(list): OMIM disorders of the query patient
        pheno_matching_patient(patient_obj): a patient object from the database
        max_similarity(float): a floating point number representing the highest value allowed for a feature

    Returns:
        patient_similarity(float): the computed phenotype similarity among the patients
    """
    patient_similarity = 0
    hpo_score = 0
    omim_score = 0

    max_omim_score = 0
    max_hpo_score = 0

    # get matching patients HPO terms as a list
    matching_hpo_terms = features_to_hpo(pheno_matching_patient.get("features"))
    matching_omim_terms = disorders_to_omim(pheno_matching_patient.get("disorders"))

    # If both query patient and matching patient contain features to compare (HPO terms)
    if hpo_terms and matching_hpo_terms:
        # If both query patient and matching patient contain OMIM diagnoses
        if disorders and matching_omim_terms:
            # LOG.info("OMIM diagnoses available for comparison")
            max_omim_score = max_similarity / 2
            max_hpo_score = max_similarity / 2

        else:  # OMIM diagnoses are missing --> HPO score represents max similarity
            max_hpo_score = max_similarity

        hpo_score = similarity_wrapper(hpoic, hpo, max_hpo_score, hpo_terms, matching_hpo_terms)

    else:  # HPO terms missing
        # similarity is computed using OMIM terms,
        # Penalty for missing HPO terms: max_omim_score = max_similarity/2
        max_omim_score = max_similarity / 2

    if max_omim_score:  # OMIM terms can be compared
        omim_score = evaluate_subcategories(disorders, matching_omim_terms, max_omim_score)

    patient_similarity = hpo_score + omim_score
    return patient_similarity


def similarity_wrapper(hpoic, hpo, max_hpo_score, hpo_terms_q, hpo_terms_m):
    """Calculate patient similarity based on HPO terms from2 patients

    Args:
        hpoic(class) : the information content for the HPO
        hpo(class): an instance of the class for interacting with HPO
        max_hpo_score(float): max score which can be assigned to HPO similarity
        hpo_terms_q(list): a list of HPO terms from query patient
        hpo_terms_m(list): a list of HPO terms from match patient

    Returns:
        score(float): simgic similarity score after HPO term comparison
    """
    # create Patient object from query patient data:
    terms = set()
    for term_id in hpo_terms_q:
        term = hpo[term_id]
        if term:
            terms.add(term)
    query_patient = Patient(pat_id="q", hp_terms=terms)

    # create Patient object from match patient data:
    terms = set()
    for term_id in hpo_terms_m:
        term = hpo[term_id]
        if term:
            terms.add(term)
    match_patient = Patient(pat_id="m", hp_terms=terms)

    # Get simgic similarity score for HPO terms comparison
    # Range is 0 to 1, with 0=no similarity and 1=highest similarity
    simgic_score = pheno_similarity_score_simgic(
        hpoic=hpoic, patient1=query_patient, patient2=match_patient
    )
    relative_simgic_score = simgic_score * max_hpo_score
    return relative_simgic_score


def evaluate_subcategories(list1, list2, max_score):
    """returns a numerical representation of the similarity of two lists of strings

    Args:
        list1(list): a list of strings (this is a list of items from the query patient)
        list2(list): another list of strings (list of items from the patients in database)
        max_score(float): the maximum value to return if the lists are identical

    Returns:
        matching_score(float): a number reflecting the similarity between the lists
    """
    matching_score = 0
    if len(list1) > 0:
        list_item_score = max_score / len(
            list1
        )  # the max value of each matching item between lists
        n_shared_items = len(
            set(list1).intersection(list2)
        )  # number of elements shared between the lists
        matching_score = n_shared_items * list_item_score
    return matching_score



# -*- coding: utf-8 -*-
import logging

import requests
from clint.textui import progress
from patientMatcher.constants import PHENOTYPE_TERMS

LOG = logging.getLogger(__name__)


def update_resources(test=True):
    """Download phenotype files necessary to perform phenotype matching"""

    for key, item in PHENOTYPE_TERMS.items():
        destination = item["resource_path"]
        url = item["url"]

        r = requests.get(url, stream=True)
        total_length = int(r.headers.get("content-length"))

        if total_length == 0:  # read file and get its size
            LOG.warning(f"Could not verify file {key} at the requested url:{url}")

        if test or total_length == 0:
            LOG.warning(f"File {key} has a length of {total_length}")
            continue

        with open(destination, "wb") as f:  # overwrite file
            for chunk in progress.bar(
                r.iter_content(chunk_size=1024), expected_size=(total_length / 1024) + 1
            ):
                if chunk:
                    f.write(chunk)
                    f.flush()
        LOG.debug(f"File {key} written to disk")



# -*- coding: utf-8 -*-
import logging

from flask import current_app, jsonify
from jsonschema import ValidationError
from patientMatcher.__version__ import __version__
from patientMatcher.auth.auth import authorize
from patientMatcher.constants import STATUS_CODES
from patientMatcher.match.handler import external_matcher
from patientMatcher.parse.patient import EMAIL_REGEX, href_validate, mme_patient, validate_api
from patientMatcher.utils.delete import delete_by_query
from patientMatcher.utils.patient import patients
from patientMatcher.utils.stats import general_metrics

LOG = logging.getLogger(__name__)


def populate_index_data():
    """Populate the dictionary used to display the server index page"""
    data = dict(
        node_stats=metrics(),
        node_disclaimer=current_app.config.get("DISCLAIMER"),
        node_contacts=current_app.config.get("ADMINS"),
        connected_nodes=len(get_nodes(current_app.db)),
        software_version=__version__,
    )
    return data


def heartbeat(disclaimer):
    """Return a heartbeat as defined here:https://github.com/ga4gh/mme-apis/blob/master/heartbeat-api.md"""

    hbeat = {
        "heartbeat": {
            "production": current_app.config.get("TESTING", "True") in ["False", False],
            "version": __version__,
            "accept": [
                "application/vnd.ga4gh.matchmaker.v1.0+json",
                "application/vnd.ga4gh.matchmaker.v1.1+json",
            ],
        },
        "disclaimer": disclaimer,
    }
    return hbeat


def metrics():
    """return database metrics"""
    db_metrics = general_metrics(current_app.db)
    return db_metrics


def get_nodes(database):
    """Get all connected nodes as a list of objects with node_id and node_label as elements"""
    results = list(database["nodes"].find())
    nodes = []
    for node in results:
        nodes.append({"id": node["_id"], "description": node["label"]})
    return nodes


def patient(database, patient_id):
    """Return a mme-like patient from database by providing its ID"""
    query_patient = None
    query_result = list(patients(database=database, ids=[patient_id]))
    if query_result:
        query_patient = query_result[0]
    return query_patient


def match_external(database, query_patient, node=None):
    """Trigger an external patient matching for a given patient object"""
    # trigger the matching and save the matching id to variable
    matching_obj = external_matcher(database, query_patient, node)
    # save matching object to database only if there are results or error messages
    if matching_obj and (matching_obj.get("has_matches") or matching_obj.get("errors")):
        database["matches"].insert_one(matching_obj)
    return matching_obj


def check_request(database, request):
    """Check if request is valid, if it is return MME formatted patient
    Otherwise return error code.
    """

    # check that request is using a valid auth token
    if not authorize(database, request):
        LOG.warning("Request is not authorized")
        return 401

    try:  # make sure request has valid json data
        request_json = request.get_json(force=True)
    except Exception as err:
        LOG.warning("Json data in request is not valid:{}".format(err))
        return 400

    try:
        # validate json data against MME API
        validate_api(json_obj=request_json, is_request=True)
    except Exception as err:
        LOG.warning("Patient data does not conform to API:{}".format(err))
        return 422

    # validate and eventually fix patient's contact href
    contact_href = request_json["patient"]["contact"]["href"]

    # If new contact is a simple email, add "mailto" schema
    if bool(EMAIL_REGEX.match(contact_href)) is True and not "mailto:" in contact_href:
        contact_href = ":".join(["mailto", contact_href])
        request_json["patient"]["contact"]["href"] = contact_href

    if href_validate(contact_href) is False:
        LOG.warning(
            "Provided contact href does not have a valid schema - either a URL (http://.., https://..) or an email address (mailto:..)"
        )
        return 422

    formatted_patient = mme_patient(json_patient=request_json["patient"], convert_to_ensembl=True)
    return formatted_patient


def validate_response(matches):
    """Validates patient matching results before sending them away in a response"""

    try:  # validate json data against MME API
        validate_api(json_obj=matches, is_request=False)
    except ValidationError as err:
        LOG.info("Patient data does not conform to API:{}".format(err))
        return 422
    return matches


def bad_request(error_code):
    """Crete an automatic response based on custom error codes"""
    message = STATUS_CODES[error_code]["message"]
    resp = jsonify(message)
    resp.status_code = error_code
    return resp


def delete_patient(database, patient_id):
    """Remove a patient by ID"""
    message = ""

    # first delete all matches in database for this patient:
    query = {"data.patient.id": patient_id}
    deleted = delete_by_query(query, database, "matches")
    LOG.info("deleted {} matche/s triggered by this patient".format(deleted))

    query = {"_id": patient_id}
    deleted = delete_by_query(query, database, "patients")
    message = {}
    if deleted == 1:
        message["message"] = "Patient and its matches were successfully deleted from database"
    else:
        message["message"] = "ERROR. Could not delete a patient with ID {} from database".format(
            patient_id
        )
    return message



# -*- coding: utf-8 -*-
import json
import logging
import re
from copy import deepcopy
from pkgutil import get_data
from urllib.parse import urlparse

from jsonschema import FormatChecker, RefResolver, validate
from patientMatcher.utils.gene import ensembl_to_symbol, entrez_to_symbol, symbol_to_ensembl
from patientMatcher.utils.variant import liftover

LOG = logging.getLogger(__name__)
SCHEMA_FILE = "api.json"
EMAIL_REGEX = re.compile(r"[^@]+@[^@]+\.[^@]+")


def href_validate(href):
    """Validates the URL used as contact href for the patient

    Args:
        href(str): A string provided by the user as patient's contact href
    Returns:
        True if it's a valid URL or False if it isn't
    """
    LOG.warning(f"Passed href is {href}")
    try:
        result = urlparse(href)
        if result.scheme == "mailto":  # mailto:me@example.com
            # validate email
            return bool(EMAIL_REGEX.match(href.split("mailto:")[1]))
        return all([result.scheme, result.netloc]) and result.scheme in ["http", "https", "mailto"]
    except Exception as ex:
        return False


def mme_patient(json_patient, convert_to_ensembl=False):
    """
    Accepts a json patient and converts it to a MME patient,
    formatted as required by patientMatcher database

    Args:
        patient_obj(dict): a patient object as in https://github.com/ga4gh/mme-apis
        convert_to_entrez(bool): convert gene IDs to ensembl IDs

    Returns:
        mme_patient(dict) : a mme patient entity
    """

    # Make sure gene objects are defined by ensembl IDs
    if json_patient.get("genomicFeatures") and convert_to_ensembl:
        format_genes(json_patient)

    mme_patient = {
        "_id": json_patient["id"],
        "id": json_patient["id"],
        "label": json_patient.get("label"),
        "sex": json_patient.get("sex"),
        "contact": json_patient["contact"],
        "features": json_patient.get("features"),
        "genomicFeatures": json_patient.get("genomicFeatures"),
        "disorders": json_patient.get("disorders"),
        "species": json_patient.get("species"),
        "ageOfOnset": json_patient.get("ageOfOnset"),
        "inheritanceMode": json_patient.get("inheritanceMode"),
    }

    # remove keys with empty values from mme_patient object
    mme_patient = {k: v for k, v in mme_patient.items() if v is not None}

    return mme_patient


def json_patient(mme_patient):
    """Converts a mme patient into a json-like as in the MME APIs

    Args:
        mme_patient(dict): a patient object as it is stored in database

    Returns:
        json_patient(dict): a patient object conforming to MME API
    """
    json_patient = mme_patient
    if "id" not in mme_patient:
        json_patient["id"] = json_patient["_id"]
    if "_id" in json_patient:
        json_patient.pop("_id")

    return json_patient


def features_to_hpo(features):
    """Extracts HPO terms from a list of phenotype features of a patient

    Args:
        features(list): a list of features dictionaries

    Returns:
        hpo_terms(list): a list of HPO terms. Example : ['HP:0100026', 'HP:0009882', 'HP:0001285']
    """
    if features is None:
        return []
    hpo_terms = [feature.get("_id") for feature in features if feature.get("_id")]
    if len(hpo_terms) == 0:
        hpo_terms = [feature.get("id") for feature in features if feature.get("id")]
    return hpo_terms


def disorders_to_omim(disorders):
    """Extracts OMIM terms from a list of disorders of a patient

    Args:
        disorders(list): a list of disorders

    Returns:
        omim_terms(list): a list of OMIM terms. Example : ['MIM:616007', 'MIM:614665']
    """
    if disorders is None:
        return []
    omim_terms = [disorder.get("id") for disorder in disorders if disorder.get("id")]
    return omim_terms


def format_genes(patient_obj):
    """Checks if patient's gFeatures gene ids are defined as ensembl ids.
    If they are entrez ids or symbols thet will be converted to ensembl IDs.

    Args:
        patient_obj(dict): A patient object with genotype features
    """
    formatted_features = []
    for feature in patient_obj.get("genomicFeatures", []):
        if "gene" in feature and feature["gene"].get("id"):
            gene = feature["gene"]["id"]
            gene_id, symbol = _convert_gene(gene)
            if symbol:
                feature["gene"]["_geneName"] = symbol  # add non-standard but informative field
            feature["gene"]["id"] = gene_id
        formatted_features.append(feature)

    if formatted_features:
        patient_obj["genomicFeatures"] = formatted_features


def _convert_gene(gene):
    """Convert provided gene id to Ensembl gene id return it with eventual HGNC gene symbol

    Args:
        gene(str): can be either be:
            - a string representing an entrez id. example: "4556"
            - a HGNC gene symbol, example: "GPX4"
            - An Ensembl gene ID, example: "ENSG00000167468"

    Returns:
        gene(str): An Ensembl gene ID, example: "ENSG00000167468"
        symbol(str): HGNC gene symbol, example: "GPX4"
    """
    symbol = None
    try:
        if gene.startswith("ENSG"):  # Ensembl gene ID
            symbol = ensembl_to_symbol(gene)
            return gene, symbol
        if gene.isdigit():  # Entrez id
            symbol = entrez_to_symbol(gene)
        else:  # HGNC gene symbol
            symbol = gene
        if symbol:
            gene = symbol_to_ensembl(symbol) or gene

    except Exception as ex:
        LOG.error(
            f"An error occurred while converting gene format using the Ensembl Rest API: {ex}"
        )

    return gene, symbol


def gtfeatures_to_genes_symbols(gtfeatures):
    """Extracts all gene names from a list of genomic features
    Args:
        gtfeatures(list): a list of genomic features objects

    Returns:
        gene_set(list): a list of unique gene names contained in the features
    """
    genes = []
    symbols = []
    for feature in gtfeatures:
        if "gene" in feature:
            gene = feature["gene"].get("id")
            gene, symbol = _convert_gene(gene)
            if gene:
                genes.append(gene)
            if symbol:
                symbols.append(symbol)

    gene_set = list(set(genes))
    symbol_set = list(set(symbols))
    return gene_set, symbol_set


def lift_variant(variant):
    """Perform a variant liftover using Ensebl REST API and return eventual variant in the other genome build

    Args:
        variant(dict): example:
            {'assembly': 'GRCh38', 'referenceName': '12', 'start': 14641142, 'end': 14641142, 'referenceBases': 'C', 'alternateBases': 'T'}

    Returns:
        lifted_variants(list of dict): example:
            [{'assembly': 'GRCh37', 'referenceName': '12', 'start': 14794076, 'end': 14794076, 'referenceBases': 'C', 'alternateBases': 'T'}]
    """
    lifted_vars = []
    mappings = liftover(
        variant.get("assembly"),
        variant.get("referenceName"),
        variant.get("start") + 1,  # coordinates are 0-based in MatchMaker
        variant.get("end") + 1 if variant.get("end") else variant.get("end"),
    )

    if mappings is None:
        return lifted_vars

    for res in mappings:
        # Create a variant which is the copy of the original variant
        lifted = deepcopy(variant)
        mapped = res["mapped"]
        # Modify coordinates of this variant according to mapping results
        lifted["assembly"] = mapped["assembly"]
        lifted["referenceName"] = mapped["seq_region_name"]
        lifted["start"] = mapped["start"] - 1  # conver back to 0-based coordinates
        if variant.get("end") is None:
            continue
        lifted["end"] = mapped["end"] - 1

        lifted_vars.append(lifted)

    return lifted_vars


def gtfeatures_to_variants(gtfeatures):
    """Extracts all variants from a list of genomic features

    Args:
        gtfeatures(list): a list of genomic features objects

    Returns:
        variants(list): a list of variants
    """
    variants = []
    for feature in gtfeatures:
        if "variant" in feature:
            variant = feature["variant"]
            if variant is None:
                continue
            # Add variant to search terms
            variants.append(feature["variant"])
            # Add also corresponding variant in another genome build (GRCh38 if original variant was GRCh37, and the other way around)
            lifted_variants = lift_variant(feature["variant"])
            if not lifted_variants:
                continue  # Variant could not be lifted to the other build
            for lifted in lifted_variants:
                variants.append(lifted)

    return variants


def validate_api(json_obj, is_request):
    """Validates a request against the MME API
       The code for the validation against the API and the API specification is taken from
       this project: https://github.com/MatchmakerExchange/reference-server

    Args:
        json_obj(dict): json request or response to validate
        is_request(bool): True if it is a request, False if it is a response

    Returns
        validated(bool): True or False
    """
    validated = True
    schema = "#/definitions/response"
    if is_request:
        schema = "#/definitions/request"

    LOG.info("Validating against SCHEMA {}".format(schema))

    # get API definitions
    schema_data = json.loads(get_data("patientMatcher.resources", SCHEMA_FILE).decode("utf-8"))
    resolver = RefResolver.from_schema(schema_data)
    format_checker = FormatChecker()
    resolver_schema = resolver.resolve_from_url(schema)
    validate(json_obj, resolver_schema, resolver=resolver, format_checker=format_checker)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import click
from flask.cli import current_app, with_appcontext
from patientMatcher.utils.delete import delete_by_query


@click.group()
def remove():
    """Remove items from database using the CLI"""
    pass


@remove.command()
@click.option(
    "-i",
    "--id",
    type=click.STRING,
    nargs=1,
    required=False,
    help="ID of the patient to be removed from database",
)
@click.option(
    "-l",
    "--label",
    type=click.STRING,
    nargs=1,
    required=False,
    help="Label of the patient to be removed from database",
)
@click.option(
    "--remove_matches/--leave_matches",
    default=False,
    help="Remove or leave on db matches triggered by patient",
)
@with_appcontext
def patient(id, label, remove_matches):
    """Removing a patient from patientMatcher providing its ID"""

    if not id and not label:
        click.echo("Error: either ID and/or label should be provided to delete a patient.")
        raise click.Abort()

    if remove_matches and not id:
        click.echo("Please provide patient ID and not label to remove all its matches.")
        raise click.Abort()

    query = {}
    if id:
        query["_id"] = id
    if label:
        query["label"] = label

    n_removed = delete_by_query(query=query, mongo_db=current_app.db, mongo_collection="patients")
    click.echo("Number of patients removed from database:{}".format(n_removed))

    if remove_matches:
        # this will remove ONLY matches where this patient was the query patient
        # NOT those where patient was among the matching results
        query = {"data.patient.id": id}
        n_removed = delete_by_query(
            query=query, mongo_db=current_app.db, mongo_collection="matches"
        )
        click.echo("Number of matches for this patient removed from database:{}".format(n_removed))


@remove.command()
@click.option(
    "-i",
    "--id",
    type=click.STRING,
    nargs=1,
    required=True,
    help="ID of the client to be removed from database",
)
@with_appcontext
def client(id):
    """Remove a client from database by providing its ID"""

    query = {"_id": id}
    n_removed = delete_by_query(query=query, mongo_db=current_app.db, mongo_collection="clients")
    click.echo("Number of clients removed from database:{}".format(n_removed))


@remove.command()
@click.option(
    "-i",
    "--id",
    type=click.STRING,
    nargs=1,
    required=True,
    help="ID of the node to be removed from database",
)
@with_appcontext
def node(id):
    """Remove a node from database by providing its ID"""

    query = {"_id": id}
    n_removed = delete_by_query(query=query, mongo_db=current_app.db, mongo_collection="nodes")
    click.echo("Number of nodes removed from database:{}".format(n_removed))



# -*- coding: utf-8 -*-
import requests
import responses
from patientMatcher.match.handler import external_matcher, internal_matcher
from patientMatcher.parse.patient import mme_patient
from patientMatcher.utils.add import backend_add_patient


@responses.activate
def test_internal_matching(mock_app, database, gpx4_patients):
    """Testing the combined matching algorithm"""

    # GIVEN a mocked Ensembl REST API:
    responses.add(
        responses.GET,
        f"https://grch37.rest.ensembl.org/xrefs/symbol/homo_sapiens/GPX4?external_db=HGNC",
        json=[{"id": "ENSG00000167468", "type": "gene"}],
        status=200,
    )

    # GIVEN a mocked Ensembl gene lookup service:
    responses.add(
        responses.GET,
        f"https://grch37.rest.ensembl.org/lookup/id/ENSG00000167468",
        json=[{"display_name": "GPX4"}],
        status=200,
    )

    # GIVEN a mocked liftover service:
    responses.add(
        responses.GET,
        f"https://grch37.rest.ensembl.org/map/human/GRCh37/19:1105813..1105814/GRCh38?content-type=application/json",
        json=[],
        status=200,
    )
    responses.add(
        responses.GET,
        f"https://grch37.rest.ensembl.org/map/human/GRCh37/19:1106232..1106238/GRCh38?content-type=application/json",
        json=[],
        status=200,
    )

    # load 2 test patients in mock database
    for patient in gpx4_patients:
        mme_pat = mme_patient(patient, True)  # convert gene symbol to ensembl
        database["patients"].insert_one(mme_pat).inserted_id

    # 2 patients should be inserted
    results = database["patients"].find({"genomicFeatures.gene.id": "ENSG00000167468"})
    assert len(list(results)) == 2

    # test matching of one of the 2 patients against both patients in database
    proband_patient = mme_patient(gpx4_patients[0], True)

    match = internal_matcher(database, proband_patient, 0.5, 0.5)
    match_patients = match["results"][0]["patients"]
    assert len(match_patients) == 2

    higest_scored_patient = match_patients[0]  # first returned patient has higher score
    lowest_scored_patient = match_patients[-1]  # last returned patient has lower score

    assert higest_scored_patient["score"]["patient"] > lowest_scored_patient["score"]["patient"]


@responses.activate
def test_internal_matching_with_threshold(mock_app, database, gpx4_patients):

    # GIVEN a mocked Ensembl REST API:
    responses.add(
        responses.GET,
        f"https://grch37.rest.ensembl.org/xrefs/symbol/homo_sapiens/GPX4?external_db=HGNC",
        json=[{"id": "ENSG00000167468", "type": "gene"}],
        status=200,
    )

    # GIVEN a mocked Ensembl gene lookup service:
    responses.add(
        responses.GET,
        f"https://grch37.rest.ensembl.org/lookup/id/ENSG00000167468",
        json=[{"display_name": "GPX4"}],
        status=200,
    )

    # GIVEN a mocked liftover service:
    responses.add(
        responses.GET,
        f"https://grch37.rest.ensembl.org/map/human/GRCh37/19:1105813..1105814/GRCh38?content-type=application/json",
        json=[],
        status=200,
    )
    responses.add(
        responses.GET,
        f"https://grch37.rest.ensembl.org/map/human/GRCh37/19:1106232..1106238/GRCh38?content-type=application/json",
        json=[],
        status=200,
    )

    # load 2 test patients in mock database
    for patient in gpx4_patients:
        mme_pat = mme_patient(patient, True)  # convert gene symbol to ensembl
        database["patients"].insert_one(mme_pat).inserted_id

    # 2 patients should be inserted
    results = database["patients"].find({"genomicFeatures.gene.id": "ENSG00000167468"})
    assert len(list(results)) == 2

    # test matching of one of the 2 patients against both patients in database
    proband_patient = mme_patient(gpx4_patients[0], True)

    match = internal_matcher(
        database=database,
        patient_obj=proband_patient,
        max_pheno_score=0.5,
        max_geno_score=0.5,
        max_results=5,
        score_threshold=0.5,
    )
    match_patients = match["results"][0]["patients"]
    assert len(match_patients) == 1  # one patient is filtered out by search threshold


def test_external_matching(database, test_node, gpx4_patients, monkeypatch):
    """Testing the function that trigger patient matching across connected nodes"""

    patient = gpx4_patients[0]

    # insert test node object in database
    database["nodes"].insert_one(test_node)

    # insert patient object in database
    inserted_ids = backend_add_patient(mongo_db=database, patient=patient, match_external=False)
    assert inserted_ids

    class MockResponse(object):
        def __init__(self):
            self.status_code = 200

        def json(self):
            resp = {"disclaimer": "This is a test disclaimer", "results": gpx4_patients}
            return resp

    def mock_response(*args, **kwargs):
        return MockResponse()

    monkeypatch.setattr(requests, "request", mock_response)

    ext_m_result = external_matcher(database, patient, test_node["_id"])
    assert isinstance(ext_m_result, dict)
    assert ext_m_result["data"]["patient"]["id"] == patient["id"]
    assert ext_m_result["has_matches"] == True
    assert ext_m_result["match_type"] == "external"



# -*- coding: utf-8 -*-

from patientMatcher.utils.add import add_node


def test_add_client(database, test_client):
    """Test adding a client with auth token to the database"""

    is_client = True

    # no clients should be present in demo database
    assert database["clients"].find_one() is None

    inserted_id, collection = add_node(mongo_db=database, obj=test_client, is_client=is_client)

    assert inserted_id == test_client["_id"]
    assert collection == "clients"
    assert database["clients"].find_one()

    # make sure that once a node is inserted it's not possible to add another node with the same id
    inserted_id, collection = add_node(mongo_db=database, obj=test_client, is_client=True)
    assert inserted_id == None


def test_add_server(database, test_node):
    """Test adding a server with auth token to the database"""

    is_client = False

    # no server should be present in demo database
    assert database["nodes"].find_one() is None

    inserted_id, collection = add_node(mongo_db=database, obj=test_node, is_client=is_client)

    assert inserted_id == test_node["_id"]
    assert collection == "nodes"
    assert database["nodes"].find_one()

    # make sure that once a node is inserted it's not possible to add another node with the same id
    inserted_id, collection = add_node(mongo_db=database, obj=test_node, is_client=is_client)
    assert inserted_id == None



import responses
from patientMatcher.cli.commands import cli
from patientMatcher.constants import PHENOTYPE_TERMS


@responses.activate
def test_update_resources(mock_app):
    """Test the command that updates the database resources (diseases and HPO terms)"""

    # Given a mocked response from the servers containing the resources to be downloaded
    for key, item in PHENOTYPE_TERMS.items():

        local_resource_path = item["resource_path"]  # Resource on the local repo
        url = item["url"]  # Resource internet URL
        with open(local_resource_path, "r") as res:
            responses.add(
                responses.GET,
                url,
                body=res.read(),
                status=200,
                content_type="application/octet-stream",
                auto_calculate_content_length=True,
                stream=True,
            )

    runner = mock_app.test_cli_runner()

    # run resources update command with --test flag:
    result = runner.invoke(cli, ["update", "resources", "--test"])
    assert result.exit_code == 0


def test_update_contact(mock_app, gpx4_patients):
    """Test the command to bulk-update patients contact"""

    runner = mock_app.test_cli_runner()
    patients_collection = mock_app.db.patients

    # GIVEN a database with some patients
    patients_collection.insert_many(gpx4_patients)
    test_patients = patients_collection.find()
    # Sharing a contact information
    contacts = test_patients.distinct("contact.href")
    assert len(contacts) == 1

    # WHEN their contact info is updated using the cli
    new_href = "new.contact@mail.com"
    result = runner.invoke(
        cli,
        [
            "update",
            "contact",
            "--old-href",
            contacts[0],
            "--href",
            new_href,
            "--name",
            "New Name",
            "--institution",
            "Test Institution",
        ],
        input="y",
    )
    assert result.exit_code == 0

    # THEN the config info should be updated
    updated_patient = patients_collection.find({"contact.href": ":".join(["mailto", new_href])})
    assert len(list(updated_patient)) > 0


def test_update_contact_no_href_match(mock_app, gpx4_patients):
    """Test the command to bulk-update patients contact when old contact href is not matching any patients"""

    runner = mock_app.test_cli_runner()
    patients_collection = mock_app.db.patients

    # GIVEN a database with some patients
    patients_collection.insert_many(gpx4_patients)
    test_patients = patients_collection.find()
    # Sharing a contact information
    contacts = test_patients.distinct("contact.href")
    assert len(contacts) == 1
    old_contact_href = contacts[0]

    # GIVEN a contact href without matches in the patients documents
    wrong_href = "some_href"
    assert wrong_href not in old_contact_href

    # WHEN their contact info is updated using the cli
    new_href = "new.contact@mail.com"
    result = runner.invoke(
        cli,
        [
            "update",
            "contact",
            "--old-href",
            wrong_href,
            "--href",
            new_href,
            "--name",
            "New Name",
            "--institution",
            "Test Institution",
        ],
    )
    assert result.exit_code == 0

    # THEN no patients contact should be updated
    assert patients_collection.find_one({"contact.href": ":".join(["mailto", new_href])}) is None


def test_update_contact_multiple_href_match(mock_app, gpx4_patients):
    """Test the command to bulk-update patients contact when old contact href is matching more than one patient contact"""

    runner = mock_app.test_cli_runner()
    patients_collection = mock_app.db.patients

    assert len(gpx4_patients) == 2
    # GIVEN a database with 2 patients with sligthly different contact href
    gpx4_patients[0]["contact"]["href"] = "test_1@mail.com"
    gpx4_patients[0]["contact"]["href"] = "test_2@mail.com"
    patients_collection.insert_many(gpx4_patients)

    # WHEN their contact info is updated using the cli but the search for the old href returns multiple contacts
    old_href = "test_"
    new_href = "test_3@mail.com"
    result = runner.invoke(
        cli,
        [
            "update",
            "contact",
            "--old-href",
            old_href,
            "--href",
            new_href,
            "--name",
            "New Name",
            "--institution",
            "Test Institution",
        ],
    )

    # THEN no patients contact should be updated
    assert patients_collection.find_one({"contact.href": ":".join(["mailto", new_href])}) is None



# -*- coding: UTF-8 -*-
import pytest
import responses
from patientMatcher.utils import ensembl_rest_client as ensembl_api
from requests.exceptions import MissingSchema


@responses.activate
def test_except_on_invalid_response():
    """Test function that creates exception with message when response returned from Ensembl REST API has status code !=200"""

    # GIVEN a response from Ensembl service with status != 200
    responses.add(
        responses.GET,
        f"https://rest.ensembl.org//info/ping",
        json={"ping": 1},
        status=500,
    )
    client = ensembl_api.EnsemblRestApiClient()

    # Then it should trigger an exception
    with pytest.raises(Exception):
        result = client.except_on_invalid_response(mockresp)


@responses.activate
def test_ping_ensembl():
    """Test ping ensembl server"""

    # GIVEN a mocked Ensembl REST API server
    responses.add(
        responses.GET,
        f"https://rest.ensembl.org//info/ping",
        json={"ping": 1},
        status=200,
    )

    client = ensembl_api.EnsemblRestApiClient(build="38")
    data = client.ping_server()
    assert data == {"ping": 1}


@responses.activate
def test_send_gene_request():
    """Test send request with correct params and endpoint"""

    # GIVEN a gene with Ensembl ID
    url = "https://grch37.rest.ensembl.org/lookup/id/ENSG00000103591"

    # GIVEN a mocked Ensembl gene lookup service:
    responses.add(
        responses.GET,
        url,
        json={"display_name": "AAGAB"},
        status=200,
    )

    client = ensembl_api.EnsemblRestApiClient()
    data = client.send_request(url)

    # get info for the ensembl gene
    assert data["display_name"] == "AAGAB"


@responses.activate
def test_send_request_wrong_url():
    """Successful requests are tested by other tests in this file.
    This test will trigger errors instead.
    """

    not_an_url = "foo"

    client = ensembl_api.EnsemblRestApiClient()
    data = client.send_request(not_an_url)
    assert type(data) == MissingSchema

    url = f"https://grch37.rest.ensembl.org/{not_an_url}"
    # GIVEN a mocked Ensembl gene lookup service:
    responses.add(
        responses.GET,
        url,
        json={"error": "page not found"},
        status=404,
    )
    data = client.send_request(url)
    assert type(data) == ValueError



import os

# Turns on debugging features in Flask
DEBUG = os.getenv("DEBUG", "True") == "True" or True

# Set to False is app if a production app
TESTING = os.getenv("TESTING", "True") == "True" or True

# secret key:
SECRET_KEY = os.getenv("SECRET_KEY") or "MySuperSecretKey"

DB_USERNAME = os.getenv("MONGODB_USER") or "pmUser"
DB_PASSWORD = os.getenv("MONGODB_PASSWORD") or "pmPassword"
DB_NAME = os.getenv("MONGODB_DBNAME") or "pmatcher"
DB_HOST = (
    os.getenv("MONGODB_HOST") or "127.0.0.1"
)  # simply substitute with 'mongodb' if connecting to MongoDB running in a container
DB_PORT = int(os.getenv("MONGODB_PORT")) if os.getenv("MONGODB_PORT") else 27017

# DB_URI = "mongodb://{}:{}@{}:{}/{}".format(DB_USERNAME, DB_PASSWORD, DB_HOST, DB_PORT, DB_NAME)
DB_URI = os.getenv("MONGODB_DB_URI") or "mongodb://{}:{}/{}".format(
    DB_HOST, DB_PORT, DB_NAME
)  # without authentication
# DB_URI = "mongodb://127.0.0.1:27011,127.0.0.1:27012,127.0.0.1:27013/?replicaSet=rs0&readPreference=primary"  # MongoDB replica set

# Matching Algorithms scores
# sum of MAX_GT_SCORE and MAX_PHENO_SCORE should be 1
MAX_GT_SCORE = float(os.getenv("MAX_GT_SCORE")) if os.getenv("MAX_GT_SCORE") else 0.75
MAX_PHENO_SCORE = float(os.getenv("MAX_PHENO_SCORE")) if os.getenv("MAX_PHENO_SCORE") else 0.25

# Max results matches returned by server.
MAX_RESULTS = int(os.getenv("MAX_RESULTS")) if os.getenv("MAX_RESULTS") else 5

# Set a minimum patient score threshold for returned results
# Set this parameter to 0 to return all results with a score higher than 0
SCORE_THRESHOLD = float(os.getenv("SCORE_THRESHOLD")) if os.getenv("SCORE_THRESHOLD") else 0

# Disclaimer. This text is returned along with match results or server metrics
DISCLAIMER = (
    os.getenv("DISCLAIMER")
    or """PatientMatcher provides data in good faith as a research tool and makes no warranty nor assumes any legal responsibility for any purpose for which the data is used. Users should not attempt in any case to identify patients whose data is returned by the service. Users who intend to publish papers using this software should acknowledge PatientMatcher and its developers (https://www.scilifelab.se/facilities/clinical-genomics-stockholm/)."""
)

# Email notification params.
# Required only if you want to send match notifications to patients contacts
MAIL_SERVER = os.getenv("MAIL_SERVER") or None  # "smtp.gmail.com"
MAIL_PORT = int(os.getenv("MAIL_PORT")) if os.getenv("MAIL_PORT") else None  # 587
MAIL_USE_TLS = os.getenv("MAIL_USE_TLS", "True") == "True" or True  # True
MAIL_USE_SSL = os.getenv("MAIL_USE_SSL", "True") == "True" or False  # False
MAIL_USERNAME = os.getenv("MAIL_USERNAME") or None  # some.email@test.se
MAIL_PASSWORD = os.getenv("MAIL_PASSWORD") or None  # email_password

# ADMINS will receive email notification is app crashes
ADMINS = os.getenv("ADMINS") or ["test.email@mail.se"]

# Set NOTIFY_COMPLETE to False if you don't want to notify variants and phenotypes by email
# This way only contact info and matching patients ID will be notified in email body
NOTIFY_COMPLETE = os.getenv("NOTIFY_COMPLETE", "False") == "True" or False



from patientMatcher.cli.commands import cli
from patientMatcher.parse.patient import mme_patient


def test_cli_remove_client(mock_app, database, test_client):

    # Add a client to database
    runner = mock_app.test_cli_runner()
    result = runner.invoke(
        cli,
        [
            "add",
            "client",
            "--id",
            test_client["_id"],
            "--token",
            test_client["auth_token"],
            "--url",
            test_client["base_url"],
        ],
    )
    assert result.exit_code == 0

    # check that the server was added to the "nodes" collection
    assert database["clients"].find_one()

    # Use the cli to remove client
    result = runner.invoke(cli, ["remove", "client", "--id", test_client["_id"]])

    # check that command is executed withour errors
    assert result.exit_code == 0

    # and that client is gone from database
    assert database["clients"].find_one() is None


def test_cli_remove_node(mock_app, database, test_node):

    # Add a node to database
    runner = mock_app.test_cli_runner()
    result = runner.invoke(
        cli,
        [
            "add",
            "node",
            "--id",
            test_node["_id"],
            "--label",
            "This is a test node",
            "--token",
            test_node["auth_token"],
            "--matching_url",
            test_node["matching_url"],
            "--accepted_content",
            test_node["accepted_content"],
        ],
    )
    assert result.exit_code == 0

    # check that the server was added to the "nodes" collection
    assert database["nodes"].find_one()

    # Use the cli to remove client
    result = runner.invoke(cli, ["remove", "node", "--id", test_node["_id"]])

    # check that command is executed withour errors
    assert result.exit_code == 0

    # and that node is gone from database
    assert database["nodes"].find_one() is None


def test_cli_remove_patient(mock_app, database, gpx4_patients, match_objs):

    runner = mock_app.test_cli_runner()

    # add a test patient to database
    test_patient = mme_patient(gpx4_patients[0], True)  # True --> convert gene symbols to ensembl
    inserted_id = mock_app.db["patients"].insert_one(test_patient).inserted_id
    assert inserted_id == gpx4_patients[0]["id"]

    # there is now 1 patient in database
    assert database["patients"].find_one()

    # test that without a valid id or label no patient is removed
    result = runner.invoke(cli, ["remove", "patient", "--id", "", "--label", ""])
    assert "Error" in result.output

    # Add mock patient matches objects to database
    database["matches"].insert_many(match_objs)
    # There should be 2 matches in database for this patient:
    results = database["matches"].find({"data.patient.id": inserted_id})
    assert len(list(results)) == 2

    # involke cli command to remove the patient by id and label
    result = runner.invoke(
        cli, ["remove", "patient", "--id", inserted_id, "--label", "350_1-test", "--leave_matches"]
    )
    assert result.exit_code == 0

    # check that the patient was removed from database
    assert database["patients"].find_one() is None

    # But matches are still there
    results = database["matches"].find({"data.patient.id": inserted_id})
    assert len(list(results)) == 2

    # Run remove patient command with option to remove matches but without patient ID
    result = runner.invoke(cli, ["remove", "patient", "--label", "350_1-test", "--remove_matches"])
    # And make sure that it doesn't work
    assert "Please provide patient ID and not label to remove all its matches." in result.output

    # Test now the proper command to remove patient matches:
    result = runner.invoke(cli, ["remove", "patient", "--id", inserted_id, "--remove_matches"])
    assert result.exit_code == 0

    # And make sure that patient removal removed its matchings
    assert database["matches"].find_one({"data.patient.id": inserted_id}) is None



# -*- coding: utf-8 -*-
import responses
from patientMatcher.utils.ensembl_rest_client import requests
from patientMatcher.utils.gene import ensembl_to_symbol, entrez_to_symbol, symbol_to_ensembl


@responses.activate
def test_entrez_to_symbol():
    """Test the function converting entrez ID to gene symbol"""
    # GIVEN an entrez ID
    entrez_id = "3735"
    # THAT should return a symbol
    symbol = "KARS"

    # GIVEN a mocked Ensembl REST API converting entrez ID to gene symbol
    responses.add(
        responses.GET,
        f"https://grch37.rest.ensembl.org/xrefs/name/human/{entrez_id}?external_db=EntrezGene",
        json=[{"display_id": symbol}],
        status=200,
    )

    # The EnsemblRestApiClient should return the right Ensembl ID
    assert entrez_to_symbol(entrez_id) == symbol


@responses.activate
def test_symbol_to_ensembl_one_ensembl_id():
    """
    Test function converting official gene symbol to ensembl ID using the Ensembl REST API
    Test case when the Ensembl API return only one Ensembl ID as result
    """

    # GIVEN a gene symbol
    hgnc_symbol = "AAGAB"
    # That should return an Ensembl ID
    ensembl_id = "ENSG00000103591"

    # GIVEN a mocked Ensembl REST API converting gene symbol to Ensembl ID
    responses.add(
        responses.GET,
        f"https://grch37.rest.ensembl.org/xrefs/symbol/homo_sapiens/{hgnc_symbol}?external_db=HGNC",
        json=[{"id": "ENSG00000103591", "type": "gene"}],
        status=200,
    )

    # The EnsemblRestApiClient should return the right Ensembl ID
    assert ensembl_id == symbol_to_ensembl(hgnc_symbol)


def test_symbol_to_ensembl_multiple_ensembl_id(monkeypatch):
    """Test function converting official gene symbol to ensembl ID using the Ensembl REST API
    Test case when the Ensembl API return multiple Ensembl gene IDs for one HGNC gene symbol
    """
    # GIVEN a gene symbol
    hgnc_symbol = "SKI"
    # Corresponfing to 2 different ensembl IDs
    rigth_ensembl_id = "ENSG00000157933"
    wrong_ensembl_id = "ENSG00000054392"

    # GIVEN a patched API response that returns data for 2 ensembl genes
    class MockResponse(object):
        def __init__(self, url):
            self.status_code = 200
            self.url = url

        def json(self):
            if hgnc_symbol in self.url:  # initial query, returns 2 Ensembl IDs
                return [
                    {"id": rigth_ensembl_id, "type": "gene"},
                    {"id": wrong_ensembl_id, "type": "gene"},
                ]
            elif (
                rigth_ensembl_id in self.url
            ):  # second call to the API, returns the HGNC info for the right gene
                return [
                    {
                        "primary_id": "HGNC:10896",
                        "display_id": "SKI",
                        "description": "SKI proto-oncogene",
                        "dbname": "HGNC",
                    }
                ]
            elif (
                wrong_ensembl_id in self.url
            ):  # second call to the API, returns the HGNC info for the wrong gene
                return [
                    {
                        "primary_id": "HGNC:18270",
                        "display_id": "HHAT",
                        "description": "hedgehog acyltransferase",
                        "dbname": "HGNC",
                    }
                ]

    def mock_get(url, headers):
        return MockResponse(url)

    monkeypatch.setattr(requests, "get", mock_get)

    # The EnsemblRestApiClient should return the right Ensembl ID
    ensembl_id = symbol_to_ensembl(hgnc_symbol)
    assert ensembl_id == rigth_ensembl_id


@responses.activate
def test_ensembl_to_symbol(monkeypatch):
    """Test converting ensembl ID to official gene symbol using the Ensembl APIs"""

    # GIVEN an Ensembl ID
    ensembl_id = "ENSG00000103591"
    # THAT should return a certain symbl
    symbol = "AAGAB"

    # GIVEN a patched API response
    class MockResponse(object):
        def __init__(self):
            self.status_code = 200

        def json(self):
            return {"display_name": symbol}

    def mock_get(url, headers):
        return MockResponse()

    monkeypatch.setattr(requests, "get", mock_get)

    # The EnsemblRestApiClient should return the right symbl
    assert ensembl_to_symbol(ensembl_id) == symbol



# -*- coding: utf-8 -*-

import logging

LOG = logging.getLogger(__name__)


class Patient(object):
    def __init__(self, pat_id, hp_terms):
        self.id = pat_id
        self.hp_terms = hp_terms
        self._ancestors = None

    def ancestors(self):
        """Return all the HPO terms ancestors for a patient"""
        if self._ancestors is None:
            ancestors = set()
            for term in self.hp_terms:
                ancestors.update(term.ancestors())
            self._ancestors = ancestors
        return self._ancestors


def pheno_similarity_score_simgic(hpoic, patient1, patient2):
    """Compare 2 patients phenotypes using the simgic algorithm"""
    LOG.info(f"Comparing patient1.id vs patient2.id")

    p1_ancestors = patient1.ancestors()
    p2_ancestors = patient2.ancestors()

    common_ancestors = p1_ancestors & p2_ancestors  # min
    all_ancestors = p1_ancestors | p2_ancestors  # max

    return hpoic.information_content(common_ancestors) / hpoic.information_content(all_ancestors)


def patients(database, ids=None, match_query=None):
    """Get all patients in the database

    Args:
        database(pymongo.database.Database)
        ids(list): a list of IDs to return only specified patients
        match_query(dict): use a query to match and retrieve patients from the database

    Returns:
        results(Iterable[dict]): list of patients from mongodb patients collection
    """
    results = None
    query = {}
    if ids:  # if only specified patients should be returned
        LOG.info("Querying patients for IDs {}".format(ids))
        query["_id"] = {"$in": ids}
    elif match_query:
        LOG.info("Return patients matching query {}".format(match_query))
        query = match_query
    else:
        LOG.info("Return all patients in database")

    results = database["patients"].find(query)
    return results



from patientMatcher.resources import path_to_hpo_terms, path_to_phenotype_annotations

# useful HTTP response status codes with messages
STATUS_CODES = {
    200: {"status_code": 200},
    400: {"message": "Invalid request JSON"},
    422: {"message": "Request does not conform to API specifications"},
    401: {"message": "Not authorized"},
    500: {"message": "An error occurred while updating the database"},
}

# phenotype terms and annotations are used by phenotype scoring algorithm and
# are updated using the CLI
PHENOTYPE_TERMS = {
    "hpo_ontology": {
        "url": "https://raw.githubusercontent.com/obophenotype/human-phenotype-ontology/master/hp.obo",
        "resource_path": path_to_hpo_terms,
    },
    "hpo_annotations": {
        "url": "https://ci.monarchinitiative.org/view/hpo/job/hpo.annotations/lastSuccessfulBuild/artifact/rare-diseases/misc/phenotype_annotation.tab",
        "resource_path": path_to_phenotype_annotations,
    },
}



# -*- coding: utf-8 -*-

import os

import pytest
from patientMatcher.parse.patient import mme_patient
from patientMatcher.utils.add import backend_add_patient, load_demo_patients
from patientMatcher.utils.delete import delete_by_query


def test_load_demo_patients(demo_data_path, database):
    """Testing if loading of 50 test patients in database is working as it should"""

    # demo data file should be present in the expected directory
    assert os.path.isfile(demo_data_path)

    # load demo data to mock database using function located under utils/load
    inserted_ids = load_demo_patients(demo_data_path, database)
    assert len(inserted_ids) == 50  # 50 test cases should be loaded

    # make sure that trying to re-insert the same patients will not work
    re_inserted_ids = load_demo_patients(demo_data_path, database)
    assert len(re_inserted_ids) == 0

    # try to call load_demo with an invalid patient file:
    inserted_ids = load_demo_patients("this_is_a_fakey_json_file.json", database)
    assert len(inserted_ids) == 0


def test_backend_remove_patient(gpx4_patients, database):
    """Test adding 2 test patients and then removing them using label or ID"""

    # test conversion to format required for the database:
    test_mme_patients = [mme_patient(json_patient=patient) for patient in gpx4_patients]

    # make sure 2 json patient are correctly parsed
    assert len(test_mme_patients) == 2

    # insert the 2 patients into the database
    inserted_ids = [
        backend_add_patient(mongo_db=database, patient=mme_patient, match_external=False)
        for mme_patient in test_mme_patients
    ]
    assert len(inserted_ids) == 2

    # make sure that inserted patients contain computed phenotypes from Monarch
    a_patient = database["patients"].find_one()
    assert a_patient

    # test removing a patient by ID:
    remove_query = {"_id": "P0001058"}
    deleted = delete_by_query(remove_query, database, "patients")
    db_patients = database["patients"].find()
    assert len(list(db_patients)) == 1

    # test removing a patient by label:
    remove_query = {"label": "350_2-test"}
    deleted = delete_by_query(remove_query, database, "patients")
    assert database["patients"].find_one() is None



# -*- coding: utf-8 -*-

import logging

from pymongo import MongoClient

LOG = logging.getLogger(__name__)


def drop_all_collections(mongo_db):
    """Deletes all collections from the database. Most likely involked when a demo instance is created.

    Args:
        mongo_db(pymongo.database.Database)
    """
    LOG.warning(f"Dropping all existing collections in database")
    for collection in mongo_db.list_collection_names():
        mongo_db[collection].drop()


def delete_by_query(query, mongo_db, mongo_collection):
    """Deletes one or more entries matching a query from a database collection

    Args:
        query(dict): query to be used for deleting the entries
        mongo_db(pymongo.database.Database)
        mongo_collection(str): the name of a collection

    Returns:
        deleted_entries(int): the number of deleted deleted entries or the error
    """

    LOG.info(
        'Removing entries from collection "{0}" using the following parameters:{1}'.format(
            mongo_collection, query
        )
    )
    deleted_entries = 0

    try:
        result = mongo_db[mongo_collection].delete_many(query)
        deleted_entries = result.deleted_count
    except Exception as err:
        deleted_entries = err

    return deleted_entries



from patientMatcher.cli.commands import cli


def test_cli_add_demo_data(mock_app, database, monkeypatch):
    """Test the class that adds demo data"""

    # GIVEN a mocked Ensembl REST API for conversion of gene symbols to Ensembl IDs
    class MockResponse(object):
        def __init__(self, url):
            self.status_code = 200
            self.gene_symbol = url.split("homo_sapiens/")[1].split("?")[0]

        def json(self):
            return [{"id": self.gene_symbol, "type": "gene"}]

    def mock_get(url, headers):
        return MockResponse(url)

    runner = mock_app.test_cli_runner()

    # make sure that "patients" collection is empty
    assert database["patients"].find_one() is None

    # run the load demo command without the -compute_phenotypes flag
    result = runner.invoke(cli, ["add", "demodata"])
    assert result.exit_code == 0

    # check that demo patients are inserted into database
    demo_patients = database["patients"].find()
    assert len(list(demo_patients)) == 50

    # check that genomic features contain genes described by HGNC gene symbols and Ensmbl IDs
    assert demo_patients[0]["genomicFeatures"][0]["gene"]["id"]
    assert demo_patients[0]["genomicFeatures"][0]["gene"]["_geneName"]

    # check that one demo client has been created
    assert database["clients"].find_one()


def test_cli_add_client(mock_app, database, test_client):

    # make sure that "clients" collection is empty
    assert database["client"].find_one() is None

    # test add a server using the app cli
    runner = mock_app.test_cli_runner()
    result = runner.invoke(
        cli,
        [
            "add",
            "client",
            "--id",
            test_client["_id"],
            "--token",
            test_client["auth_token"],
            "--url",
            test_client["base_url"],
        ],
    )

    assert result.exit_code == 0
    assert "Inserted client" in result.output

    # check that the server was added to the "nodes" collection
    assert database["clients"].find_one()

    # Try adding the client again
    result = runner.invoke(
        cli,
        [
            "add",
            "client",
            "--id",
            test_client["_id"],
            "--token",
            test_client["auth_token"],
            "--url",
            test_client["base_url"],
        ],
    )
    assert result.exit_code == 0
    # And you should get an abort message
    assert "Aborted" in result.output
    # And number of clients in database should stay the same
    results = database["clients"].find()
    assert len(list(results)) == 1


def test_cli_add_node(mock_app, database, test_node):
    # make sure that "nodes" collection is empty
    assert database["nodes"].find_one() is None

    # test add a server using the app cli
    runner = mock_app.test_cli_runner()
    result = runner.invoke(
        cli,
        [
            "add",
            "node",
            "--id",
            test_node["_id"],
            "--label",
            "This is a test node",
            "--token",
            test_node["auth_token"],
            "--matching_url",
            test_node["matching_url"],
            "--accepted_content",
            test_node["accepted_content"],
        ],
    )
    assert result.exit_code == 0
    assert "Inserted node" in result.output

    # check that the server was added to the "nodes" collection
    assert database["nodes"].find_one()

    # Try adding the node again
    result = runner.invoke(
        cli,
        [
            "add",
            "node",
            "--id",
            test_node["_id"],
            "--label",
            "This is a test node",
            "--token",
            test_node["auth_token"],
            "--matching_url",
            test_node["matching_url"],
            "--accepted_content",
            test_node["accepted_content"],
        ],
    )
    assert result.exit_code == 0
    # And you should get an abort message
    assert "Aborted" in result.output
    # And number of nodes in database should stay the same
    results = database["nodes"].find()
    assert len(list(results)) == 1



# -*- coding: utf-8 -*-
import responses
from patientMatcher.match.genotype_matcher import match
from patientMatcher.parse.patient import mme_patient


@responses.activate
def test_genotype_matching(database, gpx4_patients):
    """Testing the genotyping matching algorithm"""

    # GIVEN a mocked Ensembl REST API:
    responses.add(
        responses.GET,
        f"https://grch37.rest.ensembl.org/xrefs/symbol/homo_sapiens/GPX4?external_db=HGNC",
        json=[{"id": "ENSG00000167468", "type": "gene"}],
        status=200,
    )

    # GIVEN a mocked Ensembl gene lookup service:
    responses.add(
        responses.GET,
        f"https://grch37.rest.ensembl.org/lookup/id/ENSG00000167468",
        json=[{"display_name": "GPX4"}],
        status=200,
    )

    # GIVEN a mocked liftover service:
    responses.add(
        responses.GET,
        f"https://grch37.rest.ensembl.org/map/human/GRCh37/19:1105813..1105814/GRCh38?content-type=application/json",
        json=[],
        status=200,
    )
    responses.add(
        responses.GET,
        f"https://grch37.rest.ensembl.org/map/human/GRCh37/19:1106232..1106238/GRCh38?content-type=application/json",
        json=[],
        status=200,
    )

    # load 2 test patients in mock database
    for patient in gpx4_patients:
        mme_pat = mme_patient(patient, True)  # convert gene symbol to ensembl
        database["patients"].insert_one(mme_pat).inserted_id

    # 2 patients should be inserted
    results = database["patients"].find({"genomicFeatures.gene.id": "ENSG00000167468"})
    assert len(list(results)) == 2

    # test matching of a patient (with variants in genes) against the demo patients in database
    proband_patient = mme_patient(gpx4_patients[0], True)

    # assert patient has genomic features
    gt_features = proband_patient["genomicFeatures"]
    assert len(gt_features) == 2  # should have 2 feature to match
    assert gt_features[0]["gene"]["id"] == "ENSG00000167468"

    # match features against database with 2 patients
    matches = match(database, gt_features, 0.5)
    assert len(matches.keys()) == 2  # 2 matching patients are returned

    for key, value in matches.items():
        # patient object should also be returned
        assert "patient_obj" in value

        # genotype score for each patient should be higher than 0
        assert value["geno_score"] > 0

    # make sure that the algorithm works even if a gene or a variant object is missing:
    # remove gene ID from first gt feature
    gt_features[0]["gene"]["id"] = ""
    matches = match(database, gt_features, 0.5)
    # same patient should be returned, because of variant matching instead
    assert len(matches) == 2

    # Remove variant object from second gt feature
    gt_features[1]["variant"] = None
    matches = match(database, gt_features, 0.5)
    # same patients should be returned, because of gene matching instead
    assert len(matches.keys()) == 2
